"""
Pipeline that takes a seed dataframe, BLASTS to find sequence hits,
performs quality control, lowers alignment redundancy in a taxonomically
informed fashion, and generates an alignment.
"""

import topiary

import numpy as np
import pandas as pd

import random, string, os, shutil

def seed_to_alignment(seed_df,
                      out_dir,
                      target_seq_number=500,
                      ncbi_blast_db="nr",
                      local_blast_db=None,
                      within_species_redundancy_cutoff=0.99,
                      sparse_column_cutoff=0.95,
                      sparse_run_length_keep_percentile=0.98,
                      fx_missing_dense_cutoff=0.9,
                      align_trim=(0.05,0.95),
                      hitlist_size=5000,
                      e_value_cutoff=0.001,
                      gapcosts=(11,1),
                      num_ncbi_threads=1,
                      num_recip_blast_threads=-1,
                      overwrite=False,
                      verbose=False):
    """
    Pipeline that takes a seed dataframe, BLASTs to find sequence hits,
    performs quality control, lowers alignment redundancy in a taxonomically
    informed fashion, and generates an alignment.

    Parameters
    ----------
    seed_df : pandas.DataFrame or str
        dataframe with at least four columns: name, species, sequence,
        and aliases. See documentation on seed dataframes for details.
    out_dir : str
        output directory
    target_seq_number : int, default=500
        desired number of sequences in final alignment
    ncbi_blast_db : str or None, default="nr"
        NCBI blast database to use. If None, use a local database. Incompatible
        with local_blast_db.
    local_blast_db : str or None, default=None
        Local blast database to use. If None, use an NCBI database. Incompatible
        with ncbi_blast_db.
    within_species_redundancy_cutoff : float, default=0.99
        merge sequences with sequence identity above cutoff when removing
        redundancy of sequences within species.
    sparse_column_cutoff : float, default=0.95
        when checking alignment quality, a column is sparse if it has gaps in
        more than sparse_column_cutoff sequences.
    sparse_run_length_keep_percentile : float, 0.98
        when checking alignment quality, remove the sequences with the longest
        insertions. Toss the longest (1 - sparse_run_length_keep_percentile)*num_sequences
        sequences.
    fx_missing_dense_cutoff : float, default=0.9
        when checking alignment quality, remove sequences that are missing more
        than 1 - fx_missing_dense_cutoff of the dense (that is, not sparse,
        columns).
    align_trim : tuple, default=(0.05,0.95)
        when checking alignment quality, do not score the first and last parts
        of the alignment. Interpreted like a slice, but with percentages.
        (0.0,1.0) would not trim; (0.05,0,98) would trim the first 0.05 off the
        front and the last 0.02 off the back.
    hitlist_size : int, default=5000
        download only the top hitlist_size hits
    e_value_cutoff : float, default=0.001
        only take hits with e_value better than e_value_cutoff
    gapcost : tuple, default=(11,1)
        BLAST gapcosts (length 2 tuple of ints)
    num_ncbi_threads : int, default=1
        number of threads to use for the NCBI blast. -1 means use all available.
        (Multithreading rarely speeds up remote BLAST).
    num_recip_blast_threads : int, default=-1
        number of threads to use for local reciprocal blast. -1 means all
        available.
    overwrite : bool, default=False
        overwrite out_dir if it already exists
    verbose : bool, default=False
        verbosity of output

    Returns
    -------
    topiary_dataframe : pandas.DataFrame
        Topiary dataframe with aligned, quality-controlled sequences.
    """

    # Try to create the output directory
    out_dir = str(out_dir)
    if os.path.exists(out_dir):
        if os.path.isdir(out_dir):
            if overwrite:
                shutil.rmtree(out_dir)
            else:
                err = f"\nout_dir '{out_dir}' already exists\n\n"
                raise FileExistsError(err)
        else:
            err = f"\nout_dir '{out_dir}' already exists and is not a directory\n\n"
            raise FileExistsError(err)
    os.mkdir(out_dir)

    # If the seed_df is a file, copy that into the output directory
    if type(seed_df) is str:
        if os.path.exists(seed_df):
            root = f"00_{os.path.split(seed_df)[-1]}"
            shutil.copy(seed_df,os.path.join(out_dir,root))
            seed_df = root
        else:
            err = f"\nseed_df '{seed_df}' not found.\n\n"
            raise FileNotFoundError(err)

    # Change into the output directory
    cwd = os.getcwd()
    os.chdir(out_dir)

    step_counter = 1

    print("-------------------------------------------------------------------")
    print("Building initial topiary dataframe.")
    print("-------------------------------------------------------------------")
    print("",flush=True)

    kwargs = {"seed_df":seed_df,
              "ncbi_blast_db":ncbi_blast_db,
              "local_blast_db":local_blast_db,
              "hitlist_size":hitlist_size,
              "e_value_cutoff":e_value_cutoff,
              "gapcosts":gapcosts,
              "num_threads":num_ncbi_threads}

    out = topiary.df_from_seed(**kwargs)

    df = out[0]
    phylo_context = out[1]
    key_species = out[2]
    paralog_patterns = out[3]

    topiary.write_dataframe(df,f"{step_counter:02d}_initial-dataframe.csv")
    step_counter += 1

    print("-------------------------------------------------------------------")
    print("Doing reciprocal blast.")
    print("-------------------------------------------------------------------")
    print("",flush=True)

    proteome_list = []
    for k in key_species:
        proteome_list.append(topiary.ncbi.get_proteome(species=k))

    base = "".join([random.choice(string.ascii_letters) for _ in range(10)])
    blast_db = f"{base}_local_blast"
    topiary.ncbi.make_blast_db(proteome_list,"blast_db",overwrite=True)
    df = topiary.recip_blast(df,
                             paralog_patterns=paralog_patterns,
                             local_blast_db="blast_db",
                             num_threads=num_recip_blast_threads)

    topiary.write_dataframe(df,f"{step_counter:02d}_recip-blast-dataframe.csv")
    step_counter += 1

    print("-------------------------------------------------------------------")
    print("Reducing number of sequences.")
    print("-------------------------------------------------------------------")
    print("",flush=True)

    kwargs = {"df":df,
              "paralog_column":"recip_paralog",
              "target_seq_number":target_seq_number,
              "key_species":key_species,
              "within_species_redundancy_cutoff":within_species_redundancy_cutoff,
              "sparse_column_cutoff":sparse_column_cutoff,
              "sparse_run_length_keep_percentile":sparse_run_length_keep_percentile,
              "fx_missing_dense_cutoff":fx_missing_dense_cutoff,
              "align_trim":align_trim,
              "verbose":verbose}

    df = topiary.taxonomic_sample(**kwargs)
    topiary.write_dataframe(df,f"{step_counter:02d}_sampled-dataframe.csv")
    step_counter += 1

    print("-------------------------------------------------------------------")
    print("Aligning sequences.")
    print("-------------------------------------------------------------------")
    print("",flush=True)

    df = topiary.run_muscle(df)
    topiary.write_dataframe(df,f"{step_counter:02d}_aligned-dataframe.csv")
    step_counter += 1

    topiary.write_fasta(df,f"{step_counter:02d}_alignment.fasta",seq_column="alignment")

    os.chdir(cwd)

    return df



"""
Find the best phylogentic model to use for tree and ancestor reconstruction
given an alignment and (possibly) a tree.
"""

import topiary

from ._raxml import RAXML_BINARY, run_raxml
from topiary.external._interface import prep_calc, gen_seed, write_run_information

import pandas as pd
import numpy as np

import os, re, shutil, sys

def _parse_raxml_info_for_aic(info_file):
    """
    Open a log file from a tree evaluation run and get log likelihood,
    number of fit parameters, and various AIC scores.

    Parameters
    ----------
        info_file: raxml info file

    Return
    ------
        dictionary holding likelihood, number of free parameters, and aic score(s)
    """

    # Open the file and read lines
    out = {}
    with open(info_file,'r') as f:
        for line in f:

            # Look for likelihood
            if re.search("Final LogLikelihood:",line):
                out["L"] = float(line.strip().split(":")[-1])
                continue

            # Look for number of parameters
            if re.search("Free parameters",line):
                out["N"] = int(line.strip().split(":")[-1])
                continue

            if re.search("AIC score",line):
                cols = line.split("/")
                for c in cols:

                    value = float(c.strip().split(":")[-1])
                    key = c.strip().split(":")[0].split()[0].strip()
                    out[key] = value

    # Return L, N, and AIC
    return out

def _generate_parsimony_tree(alignment_file,
                             dir_name="parsimony-tree",
                             threads=1,
                             raxml_binary=RAXML_BINARY):
    """
    Generate a parsimony tree from an alignment.

    Parameters
    ----------
        alignment_file: alignment file in .phy format
        dir_name: name to give directory
        threads: number of threads to use
        raxml_binary: raxml binary to use

    Return
    ------
        None
    """

    run_raxml(algorithm="--start",
              alignment_file=alignment_file,
              dir_name=dir_name,
              seed=True,
              model="LG",
              threads=threads,
              raxml_binary=raxml_binary,
              log_to_stdout=False,
              other_args=["--tree","pars{1}"])

def find_best_model(df,
                    tree_file=None,
                    model_matrices=["cpREV","Dayhoff","DCMut","DEN","Blosum62",
                                    "FLU","HIVb","HIVw","JTT","JTT-DCMut","LG",
                                    "mtART","mtMAM","mtREV","mtZOA","PMB",
                                    "rtREV","stmtREV","VT","WAG","LG4M","LG4X",
                                    "PROTGTR"],
                    model_rates=["","G8"],
                    model_freqs=["","FC","FO"],
                    model_invariant=["","IO","IC"],
                    output=None,
                    overwrite=False,
                    threads=-1,
                    raxml_binary=RAXML_BINARY):
    """
    Find the best phylogentic model to use for tree and ancestor reconstruction
    given an alignment and (possibly) a tree.

    Parameters
    ----------
    df : pandas.DataFrame or str
        topiary data frame or csv written out from topiary df.
    tree_file : str
        tree_file in newick format. If not specified, parsimony tree is
        generated and used
    model_matrices : list, optional
        list of model matrices to check
    model_rates : list, optional
        ways to treat model rates
    model_freqs : list, optional
        ways to treat model freqs.
    output : str, optional
        output directory. If not specified, create an output directory with
        form "find_best_model_randomletters"
    overwrite : bool, default=False
        whether or not to overwrite existing output
    threads : int, default=-1
        number of threads to use. if -1 use all available
    raxml_binary : str, optional
        raxml binary to use

    Returns
    -------
    None
    """

    # Copy files in, write out alignment, move into working directory, etc.
    result = prep_calc(df=df,
                       tree_file=tree_file,
                       output=output,
                       overwrite=overwrite,
                       output_base="find_best_model")

    df = result["df"]
    csv_file = result["csv_file"]
    alignment_file = result["alignment_file"]
    tree_file = result["tree_file"]
    starting_dir = result["starting_dir"]

    # Generate a parsimony tree if not was specified
    if tree_file is None:

        print("\nGenerating maximum parsimony tree.\n",flush=True)

        _generate_parsimony_tree(alignment_file,
                                 dir_name="working",
                                 threads=threads,
                                 raxml_binary=raxml_binary)
        tree_file = os.path.join("working",
                                 "alignment.phy.raxml.startTree")

    # Dictionary to hold stats for each model
    out = {"model":[]}

    seed = gen_seed()

    # All possible models, dropping rate, freq, invariant for LG4M and LG4X.
    num_models = len(model_matrices)*len(model_rates)*len(model_freqs)*len(model_invariant)

    print(f"\nTrying {num_models} combinations of matrix and model parameters.\n",flush=True)
    sys.stdout.flush()

    # Go over all combos of the requested matrices, rates, and freqs.
    model_counter = 1
    for matrix in model_matrices:
        for rate in model_rates:
            for freq in model_freqs:
                for invariant in model_invariant:

                    # Build model string (for example: LG+G8+FC+IO)
                    model = [matrix,rate,freq,invariant]
                    model = [m for m in model if m != ""]
                    model = "+".join(model)

                    # Print model number we're trying
                    print(f"{model} ({model_counter}/{num_models})",flush=True)
                    model_counter += 1

                    # Check for incompatible matrix/freq/rate combos
                    if matrix in ["LG4M","LG4X"]:
                        if rate != "" or freq != "" or invariant != "":
                            print(f"skpping incompatible model combination {model}",flush=True)
                            continue

                    # Optimize branch lengths etc. on the existing tree. The
                    # --force model_lh_impr allows calculation to proceed even
                    # if this particular model didn't optimize well.
                    run_raxml(algorithm="--evaluate",
                              alignment_file=alignment_file,
                              tree_file=tree_file,
                              model=model,
                              seed=seed,
                              dir_name="tmp",
                              threads=threads,
                              raxml_binary=raxml_binary,
                              log_to_stdout=False,
                              other_args=["--force","model_lh_impr"])

                    # Grab the info file from this run
                    os.chdir("tmp")

                    # Get results from the info file
                    result = _parse_raxml_info_for_aic("alignment.phy.raxml.log")
                    out["model"].append(model)
                    for r in result:
                        try:
                            out[r].append(result[r])
                        except KeyError:
                            out[r] = [result[r]]

                    # Get out of temporary directory and nuke
                    os.chdir("..")
                    shutil.rmtree("tmp")

    # Create a dataframe sorted best to worst aicc
    final_df = pd.DataFrame(out)

    min_aic = np.min(final_df.AICc)
    final_df["p"] = np.exp((min_aic - final_df.AICc)/2)
    final_df["p"] = final_df["p"]/np.sum(final_df["p"])
    indexer = np.argsort(final_df.p)[::-1]
    final_df = final_df.iloc[indexer,:]

    # All output goes into the output directory
    outdir = "output"
    os.mkdir(outdir)

    # Write run information
    write_run_information(outdir=outdir,
                          df=df,
                          calc_type="find_best_model",
                          model=final_df.model.iloc[0],
                          cmd=None)

    # Write dataframe comparing models
    final_df.to_csv(os.path.join(outdir,"model-comparison.csv"))

    # Print best model to stdout
    print("\nTop 10 models:\n")
    print(f"{'model':>20s}{'AICc prob':>20s}",flush=True)
    for i in range(10):
        if i >= len(final_df):
            break
        print(f"{final_df.model.iloc[i]:>20s}{final_df.p.iloc[i]:>20.3f}",flush=True)

    print(f"\nWrote results to {os.path.abspath(outdir)}\n")

    # Leave the output directory
    os.chdir(starting_dir)



"""
Draw tree output from a maximum likelihood tree calculation, gene/species tree
reconcilation, or ancestral sequence reconstruction.
"""

import topiary
from topiary.external._interface import read_previous_run_dir
import inspect

def tree(run_dir,
         output_file=None,
         tip_columns=["species","nickname"],
         tip_name_separator="|",
         fontsize=36,
         circle_radius=0.025,
         value_range=None,
         R=(0.95,0),G=(0.95,0),B=(1,1),
         event_color_map={"D":"green","L":"red","S":"blue","T":"pink"},
         width=800,space_between_taxa=10,line_width=4,
         df=None):
    """
    Draw tree output from an ML tree calculation, gene/species tree
    reconcilation, or ancestral sequence reconstruction. Plot type is determined
    based on the type of calculation passed in as run_dir.

    Parameters
    ----------
    run_dir : str
        directory containing an ancestral reconstruction run
    output_file : str, optional
        output file to write tree. If running in a notebook, will return to
        notebook and write to this file. If running in a notebook but not
        specified, do not write to a file. If not in a notebook and not
        specified, will write out ancestor-tree.pdf.
    tip_columns: list, default=["species","nickname"]
        label the tree tips as "|".join(tip_columns). For example, if
        tip_columns is ["species","nickname"], tips will have names like
        'Homo sapiens|LY96'.
    tip_name_separator : str, default="|"
        string to separate columns in tip names ("|" in tip_columns example
        above.)
    fontsize : float, default=36
        fontsize in points for labels
    circle_radius : float, default=0.025
        circle size for internal nodes (fraction of total width)
    value_range : tuple,None default=(0.6,1)
        tuple holding minimum and maximum values for support color map.
        (0,100) means miminum support value is 0, maximum support value
        is 100. If None, disable node color map.
    R : tuple, default=(0.95,0)
        tuple holding red channel values for lowest and highest supports.
    G : tuple, default=(0.95,0)
        tuple holding green channel values for lowest and highest supports.
    B : tuple, default=(1.0,1.0)
        tuple holding blue channel values for lowest and highest supports.
    event_color_map : dict
        map between evolutionary events (D,L,S,T) and named colors.
    width : int, default=800
        width of total tree in pixels
    space_between_taxa : int, default=10
        number of pixels between taxa, sets height.
    line_width : int, default=4
        width of lines used to draw tree (pixels)
    df : pandas.DataFrame or None, default=None
        topiary dataframe (overides whatever is in run_dir/output/dataframe.csv)

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image; otherwise, return None
    """

    prev_run = read_previous_run_dir(run_dir)
    calc_type = prev_run["calc_type"]

    if calc_type == "ml_tree":

        # If no value_range specified, get the default from ml_tree
        if value_range is None:
            sig = inspect.signature(topiary.draw.ml_tree)
            value_range = sig.parameters["value_range"].default

        return topiary.draw.ml_tree(run_dir=run_dir,
                                    output_file=output_file,
                                    tip_columns=tip_columns,
                                    tip_name_separator=tip_name_separator,
                                    fontsize=fontsize,
                                    circle_radius=circle_radius,
                                    value_range=value_range,
                                    R=R,G=G,B=B,
                                    width=width,
                                    space_between_taxa=space_between_taxa,
                                    line_width=line_width,
                                    df=df)

    elif calc_type == "reconciliation":

        return topiary.draw.reconciliation_tree(run_dir=run_dir,
                                                output_file=output_file,
                                                tip_columns=tip_columns,
                                                tip_name_separator=tip_name_separator,
                                                fontsize=fontsize,
                                                circle_radius=circle_radius,
                                                event_color_map=event_color_map,
                                                width=width,
                                                space_between_taxa=space_between_taxa,
                                                line_width=line_width,
                                                df=df)

    elif calc_type == "ancestors":

        # If no value_range specified, get the default from ml_tree
        if value_range is None:
            sig = inspect.signature(topiary.draw.ancestor_tree)
            value_range = sig.parameters["value_range"].default

        return topiary.draw.ancestor_tree(run_dir=run_dir,
                                          output_file=output_file,
                                          tip_columns=tip_columns,
                                          tip_name_separator=tip_name_separator,
                                          fontsize=fontsize,
                                          circle_radius=circle_radius,
                                          value_range=value_range,
                                          R=R,G=G,B=B,
                                          width=width,
                                          space_between_taxa=space_between_taxa,
                                          line_width=line_width,
                                          df=df)

    else:
        err = f"\nCould not draw a tree for a calculation of type '{calc_type}'\n\n"
        raise ValueError(err)



"""
Create a nickname column that has a friendly nickname for each sequence.
"""

from topiary import _private, check

import pandas as pd
import numpy as np

import re, sys, os, string, random, pickle, io, urllib, http, copy

def create_nicknames(df,
                     paralog_patterns,
                     source_column="name",
                     output_column="nickname",
                     separator="/",
                     unassigned_name="unassigned",
                     overwrite_output=False,
                     ignorecase=True):
    """
    Create a nickname column that has a friendly nickname for each sequence,
    generated by looking for patterns defined in the :code:`paralog_patterns`
    dictionary in :code:`source_column` column from the dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe
    paralog_patterns : dict
        dictionary for creating standardized nicknames from input names. Key
        specifies what should be output, values the a list of patterns that map
        back to that key. For example:

        .. code-block:: python

            {"S100A9":["S100-A9","S100 A9","S-100 A9","MRP14"],
             "S100A8":["S100-A8","S100 A8","S-100 A9","MRP8"]}

        would assign "S100A9" to any sequence matching patterns only from its
        list; "S100A8" to any sequence matching patterns only from its list; and
        S100A9/S100A8 to any sequence matching patterns from both lists.
    source_column : str, default="name"
        source column in dataframe to use to generate a nickname
    output_column : str, default="nickname"
        column in which to store newly constructed nicknames
    separator : str, default="/"
        character to place between nicknames if more than one pattern matches.
    unassigned_name : str, default="unassigned"
        nickname to give sequences that do not match any of the patterns.
    overwrite_output : bool, default=False
        overwrite an existing output column
    ignorecase: bool, default=True
        Whether or not to ignore the case of matches when assigning the nickname.

    Returns
    -------
    topiary_dataframe : pandas.DataFrame
        Copy of dataframe with new nickname column
    """

    # Parse 'df' argument
    if type(df) is not pd.DataFrame:
        err = "\ndf should be a pandas dataframe\n\n"
        raise ValueError(err)

    # Check to make sure the specified source column exists
    try:
        df.loc[:,source_column]
    except KeyError:
        err = f"\ndataframe does not have source_column '{source_column}'\n\n"
        raise ValueError(err)

    # Make sure the output column is not reserved by topiary
    if output_column in _private.reserved_columns:
        err = f"\n'{output_column}' is a reserved column name. Please choose\n"
        err += "another column name.\n\n"
        raise ValueError(err)

    # Make sure the output_column does not exist or that we're allowed to
    # overwrite
    try:
        df.loc[:,output_column]
        if not overwrite_output:
            err = f"\ndataframe already has output_column '{output_column}'.\n"
            err += "To overwrite set overwrite_output = True\n\n"
            raise ValueError(err)
    except KeyError:
        pass

    # check unassigned_name argument
    if type(unassigned_name) is not str:
        err = "\nunassigned_name should be a string.\n\n"
        raise ValueError(err)

    # check separator argument
    if type(separator) is not str:
        err = "\nseparator should be a string.\n\n"
        raise ValueError(err)

    patterns = check.check_paralog_patterns(paralog_patterns,
                                                        ignorecase=ignorecase)

    # Get entries from source column
    source = [entry for entry in df.loc[:,source_column]]
    out = []
    for i, s in enumerate(source):

        out.append([])

        # Go over every patterns
        for p in patterns:

            # If we find the pattern in the entry, break; we only need
            # to note we found it once
            if p[0].search(source[i]):
                out[-1].append(p[1])
                break

        # Join list of hits
        out[-1] = f"{separator}".join(out[-1])

        # No hit, give unassigned_name
        if out[-1] == "":
            out[-1] = unassigned_name

    # Return an edited copy of the dataframe
    df = df.copy()
    df.loc[:,output_column] = out

    # Validate topiary dataframe to make sure not mangled; will also update
    # column order so nickname is early and thus in a user-friendly place

    return check.check_topiary_dataframe(df)



"""
Interface to ete3 for drawing phylogenetic trees. 
"""

from .tree import tree
from .ml import ml_tree
from .ancestor import ancestor_tree
from .reconciliation import reconciliation_tree
from .species import species_tree



"""
Reconcile a gene tree with a species tree using generax.
"""

import topiary
from topiary.external._interface import prep_calc, write_run_information

from ._generax import setup_generax, run_generax, GENERAX_BINARY

import ete3
import numpy as np

import os, glob, shutil

def reconcile(previous_dir=None,
              df=None,
              model=None,
              tree_file=None,
              allow_horizontal_transfer=False,
              output=None,
              overwrite=False,
              generax_binary=GENERAX_BINARY):
    """
    Reoncile the gene tree to the species tree using generax.

    Parameters
    ----------
    previous_dir : str, optional
        directory containing previous calculation. function will grab the the
        csv, model, and tree from the previous run. If this is not specified,
        `df`, `model`, and `tree_file` arguments must be specified.
    df : pandas.DataFrame or str, optional
        topiary data frame or csv written out from topiary df. Will override
        dataframe from `previous_dir` if specified.
    model : str, optional
        model (i.e. "LG+G8"). Will override model from `previous_dir`
        if specified.
    tree_file : str
        tree_file in newick format. Will override tree from `previous_dir` if
        specified.
    allow_horizontal_transfer : bool, default=False
        whether to allow horizontal transfer during reconcilation. If True, use
        the "UndatedDTL" model. If False, use the "UndatedDL" model.
    output: str, optional
        output directory. If not specified, create an output directory with
        form "generax_reconcilation_randomletters"
    overwrite : bool, default=False
        whether or not to overwrite existing output directory
    generax_binary : str, optional
        what generax binary to use

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image showing reconciled tree;
        otherwise, return None.
    """

    # Prepare for the calculation, loading in previous calculation and
    # combining with arguments as passed in.
    result = prep_calc(previous_dir=previous_dir,
                       df=df,
                       model=model,
                       tree_file=model,
                       output=output,
                       overwrite=overwrite,
                       output_base="generax_reconcilation")

    df = result["df"]
    csv_file = result["csv_file"]
    model = result["model"]
    tree_file = result["tree_file"]
    alignment_file = result["alignment_file"]
    starting_dir = result["starting_dir"]

    required = [df,model,tree_file]
    for r in required:
        if r is None:
            err = "\nA dataframe, model, and tree are required for this "
            err += "calculation.\n\n"
            raise ValueError(err)

    # Set up generax directory
    setup_generax(df,tree_file,model,"working")

    # Actually run generax
    cmd = run_generax(run_directory="working",
                      allow_horizontal_transfer=allow_horizontal_transfer,
                      generax_binary=generax_binary)

    # Make output directory to hold final outputs
    outdir = "output"
    os.mkdir(outdir)

    # Copy in tree.newick
    shutil.copy(os.path.join("working","result","results","reconcile","geneTree.newick"),
                os.path.join("output","tree.newick"))


    # Get outgroups (e.g. leaves descending from each half after root)
    reconcile_file = os.path.join("working","result","reconciliations","reconcile_events.newick")
    reconcile_tree = ete3.Tree(reconcile_file,format=1)
    root = reconcile_tree.get_tree_root()
    root_children = root.get_children()
    outgroup = [[n.name for n in r.get_leaves()] for r in root_children]

    # Write run information
    write_run_information(outdir=outdir,
                          df=df,
                          calc_type="reconciliation",
                          model=model,
                          cmd=cmd,
                          outgroup=outgroup)

    # Copy reconcilation information
    shutil.copytree(os.path.join("working","result","reconciliations"),
                    os.path.join("output","reconcilations"))

    print(f"\nWrote results to {os.path.abspath(outdir)}\n")

    # Leave working directory
    os.chdir(starting_dir)

    # Write out a summary tree.
    ret = topiary.draw.reconciliation_tree(run_dir=output,
                                           output_file=os.path.join(output,
                                                                    "output",
                                                                    "summary-tree.pdf"))
    if topiary._in_notebook:
        return ret



"""
Draw a reconciliation tree, labeling the internal nodes according to
whether they are speciations, duplications, or losses.
"""

import topiary
from topiary.external._interface import read_previous_run_dir

import os

from ._base import load_trees, create_name_dict, setup_generic_tree_format, final_render
import ete3

def reconciliation_tree(run_dir,
                        output_file=None,
                        tip_columns=["species","nickname"],
                        tip_name_separator="|",
                        fontsize=36,
                        circle_radius=0.025,
                        event_color_map={"D":"green","L":"red","S":"blue","T":"pink"},
                        width=800,space_between_taxa=10,line_width=4,
                        df=None):
    """
    Draw a reconciliation tree, labeling the internal nodes according to
    whether they are speciations, duplications, or losses.

    Parameters
    ----------
    run_dir : str
        directory containing an ancestral reconstruction run
    output_file : str, optional
        output file to write tree. If running in a notebook, will return to
        notebook and write to this file. If running in a notebook but not
        specified, do not write to a file. If not in a notebook and not
        specified, will write out ancestor-tree.pdf.
    tip_columns: list, default=["species","nickname"]
        label the tree tips as "|".join(tip_columns). For example, if
        tip_columns is ["species","nickname"], tips will have names like
        'Homo sapiens|LY96'.
    tip_name_separator : str, default="|"
        string to separate columns in tip names ("|" in tip_columns example
        above.)
    fontsize : float, default=36
        fontsize in points for labels
    circle_radius : float, default=0.025
        circle size for internal nodes (fraction of total width)
    event_color_map : dict
        map between evolutionary events (D,L,S,T) and named colors.
    width : int, default=800
        width of total tree in pixels
    space_between_taxa : int, default=10
        number of pixels between taxa, sets height.
    line_width : int, default=4
        width of lines used to draw tree (pixels)
    df : pandas.DataFrame or None, default=None
        topiary dataframe (overides whatever is in run_dir/output/dataframe.csv)

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image; otherwise, return None
    """

    # Load data from previous run
    prev_run = read_previous_run_dir(run_dir)

    # Load the tree and relevant information from the ml_run_dir output
    # directory.
    T = load_trees(prev_run_dir=run_dir,
                   tree_base_path=os.path.join("output","reconcilations"),
                   tree_names=["reconcile_events.newick"],
                   tree_fmt=[1])

    # If df not specified, get from the previous run
    if df is None:
        df = prev_run["df"]

    # Create dictionary mapping between uid and pretty name format
    name_dict = create_name_dict(df=df,
                                 tip_columns=tip_columns,
                                 separator=tip_name_separator)

    # Set up formats (color map for main nodes, tree format (ts) and generic
    # node format (ns)).
    cm, ts, ns = setup_generic_tree_format(num_leaves=len(T.get_leaves()),
                                           value_range=None,
                                           width=width,
                                           space_between_taxa=space_between_taxa,
                                           line_width=line_width)

    # Figure out if the tree has supports. If the tree did not have supports,
    # ete3 will load it in with a support value of 1.0 for all nodes. This would
    # be very unlikely for a tree with real supports, so use that to look for
    # trees without supports
    supports = []
    for n in T.traverse():
        if not n.is_leaf():
            supports.append(n.support)

    supports_seen = list(set(supports))
    if len(supports_seen) == 1 and supports_seen[0] == 1:
        has_supports = False
    else:
        has_supports = True

    circle_radius = circle_radius*width

    # Traverse T1 and T2 simultaneously
    for n in T.traverse():

        # Set base style on node of T1
        n.set_style(ns)

        # If this is not a leaf
        if not n.is_leaf():

            try:
                rgb = event_color_map[n.name]
            except KeyError:
                rgb = "gray"

            # Draw circles
            node_circle = ete3.CircleFace(radius=circle_radius,color=rgb,style="circle")
            node_circle.margin_right=-circle_radius
            n.add_face(node_circle,0,position="branch-right")

            if has_supports:

                # Wwrite out upport labels
                anc_label = ete3.TextFace(text=f"{int(round(n.support,0)):d}",fsize=fontsize)
                anc_label.inner_background.color = "white"
                anc_label.margin_right = circle_radius
                anc_label.margin_top = 0
                n.add_face(anc_label,0,position="float")

        # If this is a leaf
        else:

            # Get the clean name (rather than uid)
            clean_name = name_dict[n.name]

            # Add text for clean name
            n.name = ""
            txt = ete3.TextFace(clean_name,fsize=fontsize)
            txt.margin_left = 4
            n.add_face(txt,0,position="branch-right")

    return final_render(T,ts,output_file,"reconcilation-tree.pdf")



"""
Basic interface to opentree.
"""

import topiary
from topiary import check

from opentree import OT, taxonomy_helpers
import dendropy as dp
import ete3

import pandas as pd
import numpy as np

import re, copy

def is_allowed_phylo_context(phylo_context):
    """
    See if a phylo_context is recognizable in opentree database.

    Parameters
    ----------
    phylo_context : str
        string holding query phylognetic context

    Returns
    -------
    phylo_context : str
        validated phylo_context as a string
    """

    phylo_context = str(phylo_context)

    # Make sure the phylo_context can be recognized by the by OT
    allowed_context = OT.tnrs_contexts().response_dict
    all_allowed = []
    for k in allowed_context:
        all_allowed.extend(allowed_context[k])

    if phylo_context not in all_allowed:
        err = f"\n\nphylo_context '{phylo_context}' not recognized. Should be one of:\n\n"
        for a in all_allowed:
            err += f"    {a}\n"
        err += "\n\n"
        raise ValueError(err)

    return phylo_context

def get_phylo_context(species_list):
    """
    Get phylogenetic context (i.e. 'All life', 'Mammals') given a list of
    species names.

    Parameters
    ----------
    species_list : list
        list of binomial species names as strings (i.e. ["Homo sapiens",
        "Mus musculus"])

    Returns
    -------
    phylo_context : str
        string giving phylogenetic context
    """

    # Clean up species list
    species_list = check.check_iter(species_list,
                                    "species_list",
                                    required_value_type=str)

    # No species, return all life
    if len(species_list) == 0:
        return "All life"

    # Make sure we can get an ott for all sequences in sequence list
    ott, not_resolved = species_to_ott(species_list,phylo_context="All life")
    for k in ott:
        if ott[k][0] is None:
            err = f"\nCould not find species '{k}' in opentree database.\n"
            err += "We cannot reliably find phylogenetic context as a result.\n\n"
            raise ValueError(err)

    # Infer the phylogenetic context from this set of species
    w = OT.tnrs_infer_context(species_list)

    # Return context
    return w.response_dict["context_name"]

def species_to_ott(species_list,phylo_context="All life"):
    """
    Get the OTT values for a list of species.

    Parameters
    ----------
    species_list : list
        list of binomial species names as strings (i.e. ["Homo sapiens",
        "Mus musculus"])
    phylo_context : str, default="All life"
        used to limit species seach for looking up species ids on open tree of
        life

    Returns
    -------
    results : list
        list of ott found from this search
    not_resolved : list
        list of ott that could not be resolved on the species tree even though
        they have ott.
    """

    # Check input arguments
    species_list = check.check_iter(species_list,
                                                "species_list",
                                                required_value_type=str)

    # Make sure the phylo_context is recognizable by OTT
    phylo_context = is_allowed_phylo_context(phylo_context)

    # Strip leading/trailing spaces
    species_list = [s.strip() for s in species_list]

    # Do fuzzy match for species names
    w = OT.tnrs_match(species_list,
                      context_name=phylo_context,
                      do_approximate_matching=True)

    # Go through hits
    results = {}
    for match in w.response_dict["results"]:

        # No match
        if len(match["matches"]) == 0:
            results[match['name']] = (None,match['name'])

        # Match
        else:

            matches = match["matches"]

            # Multiple match -- print warning
            if len(matches) > 1:

                # If we have multiple fuzzy matches, where the first is not exactly
                # equal to the query, warn the user
                if matches[0]['matched_name'] != matches[0]['taxon']['unique_name']:

                    w = f"\n\nSpecies {matches[0]['matched_name']} had multiple hits. Taking first.\n"
                    w += "Matches:\n"
                    for i in range(len(matches)):
                        w += f"    {matches[i]['taxon']['unique_name']}\n"
                    print(w)

            # Record data about hit
            hit = matches[0]

            matched_name = hit["matched_name"]
            otl_name = hit["taxon"]["unique_name"]
            ott_id = hit["taxon"]["ott_id"]

            results[matched_name] = (ott_id,otl_name)


    # Record non-hits
    for unmatch in w.response_dict["unmatched_names"]:
        results[unmatch] = (None,unmatch,None)

    # List of ott numbers to query as tree
    tmp_ott_list = [results[k][0] for k in results if results[k][0] is not None]

    # Don't try to get synthetic tree for empty sequence set
    if len(tmp_ott_list) == 0:
        return results, []

    # Check for ott that cannot be resolved into trees
    ret = taxonomy_helpers.labelled_induced_synth(ott_ids=tmp_ott_list,
                                                  label_format="name_and_id",
                                                  inc_unlabelled_mrca=False)

    # Taxa unresolvable in synthetic tree
    not_resolved = []
    for bad in ret["unknown_ids"].keys():
        ott_id = ret["unknown_ids"][bad]["ott_id"]
        not_resolved.append(ott_id)

    return results, not_resolved



"""
Get a species tree given a topiary dataframe.
"""

import topiary
from topiary import check

from opentree import OT, taxonomy_helpers
import dendropy as dp
import ete3

import pandas as pd
import numpy as np

import re, copy

def get_species_tree(df):
    """
    Return an ete3 cladogram of species in tree. The leaves on the tree will
    have the following features:

    + `leaf.name`: bionomial species name as string
    + `leaf.ott`: ott as string
    + `leaf.species`: bionomial species name as string
    + `leaf.uid`: list of all uid that have this species

    Parameters
    ----------
    df : pandas.DataFrame
        topiary dataframe that has an ott column with Open Tree of Life taxon
        ids

    Returns
    -------
    species_tree : ete3.Tree
        An ete3 tree with branch lengths of 1, supports of 1, and only
        tip labels. Note: any polytomies are arbirarily resolved.
    """

    # Make sure this is a clean topiary dataframe
    df = check.check_topiary_dataframe(df)
    if "ott" not in df.columns:
        err = "\ndataframe must contain an ott column. This can be generated\n"
        err += "using topiary.opentree.get_ott_id\n\n"
        raise ValueError(err)

    # Only get keep = True
    df = df.loc[df.keep,:].copy()

    # ott_to_df_columns will be a dictionary of dictionaries. The top-level
    # dictionary is keyed to different columns. Each of
    # those keys maps to a dictionary mapping ott to a tuple of values in that
    # column that have this ott. These will be attached as features to the
    # leaf nodes in the final ete3 tree.
    to_get = ["species","ott","uid"]
    ott_to_df_columns = dict([(k,{}) for k in to_get])
    for k in to_get:

        # If the dataframe doesn't have the relevant column, do not grab
        if k not in df.columns:
            ott_to_df_columns.pop(k)
            continue

        # Load values from the dataframe into the dictionary
        for i in range(len(df)):
            row = df.iloc[i]
            try:
                ott_to_df_columns[k][row.ott].append(row[k])
            except KeyError:
                ott_to_df_columns[k][row.ott] = [row[k]]

        # Convert to tuple for each of these, as ete3 requires features are
        # hashable types.
        for o in ott_to_df_columns[k]:
            ott_to_df_columns[k][o] = tuple(ott_to_df_columns[k][o])

    # Get only rows with unique ott
    df = df.loc[df.loc[:,"ott"].drop_duplicates().index,:]

    # Make sure every species has an ott
    if np.sum(pd.isnull(df.loc[:,"ott"])) > 0:
        err = "\nNot all species have ott in the dataframe.\n"
        raise ValueError(err)

    ott_ids = list(df.loc[:,"ott"])
    species = list(df.loc[:,"species"])
    ott_species_dict = {}
    for i in range(len(ott_ids)):
        ott_species_dict[ott_ids[i]] = species[i]

    ott_as_int = [o[3:] for o in ott_ids]
    ret = taxonomy_helpers.labelled_induced_synth(ott_ids=ott_as_int,
                                                  label_format="name_and_id",
                                                  inc_unlabelled_mrca=False)

    # Write out without all the ancestor junk returned by opentree
    t = ret["labelled_tree"].as_string(schema="newick",
                                       suppress_leaf_taxon_labels=False,
                                       suppress_leaf_node_labels=True,
                                       suppress_internal_taxon_labels=True,
                                       suppress_internal_node_labels=True,
                                       suppress_edge_lengths=True,
                                       suppress_rooting=False,
                                       suppress_annotations=True,
                                       suppress_item_comments=True)

    # Read in new tree
    stripped_tree = dp.Tree.get(data=t,schema="newick")
    taxon_names = [s.label for s in stripped_tree.taxon_namespace]

    # Extract tree with labels.  This will remove all of the empty singleton
    # entries that otl brings down.
    clean_tree = stripped_tree.extract_tree_with_taxa_labels(taxon_names)

    # Rename labels on tree so they are ott
    for n in clean_tree.taxon_namespace:
        label = n.label.split("ott")[-1]
        n.label = f"ott{label}"

    # Convert tree to ete3 tree.
    final_tree = ete3.Tree(clean_tree.as_string(schema="newick",
                                                suppress_rooting=True),
                           format=9)

    # Arbitrarily resolve any polytomies
    final_tree.resolve_polytomy()

    # Give every node a support of 1 and a branch length of 1
    for n in final_tree.traverse():
        if n.dist != 1:
            n.dist = 1
        if n.support != 1:
            n.support = 1

        # Give leaves species name as a feature
        if n.is_leaf():

            name_key = str(copy.deepcopy(n.name))

            n.add_feature("species",ott_to_df_columns["species"][name_key][0])
            n.add_feature("ott",ott_to_df_columns["ott"][name_key][0])
            n.add_feature("uid",ott_to_df_columns["uid"][name_key])

            n.name = copy.deepcopy(n.species)

    return final_tree



"""
Private utility functions that are not publicly exposed in the API.
"""

import string, random, sys, inspect, argparse

required_columns = ["species","name","sequence"]
reserved_columns = required_columns[:]
reserved_columns.extend(["uid","ott","alignment","keep","always_keep"])

# Data going into a newick tree can't have any of these symbols. We also reserve
# the '#' character for comments.
reserved_characters = ["(",")",";","#",":",",","'","\""]

def generate_uid(number=1):
    """
    Generate a unique uid. This will be a 10 character random combination of
    ascii letters.

    Parameters
    ----------
        number: number of uid to generate. if 1, return a single uid. if > 1,
                return a list of uid.

    Return
    ------
        uid or list of uid
    """

    if number < 1:
        err = "number must be 1 or more\n"
        raise ValueError(err)

    out = []
    for n in range(number):
        out.append("".join([random.choice(string.ascii_letters)
                            for _ in range(10)]))

    if number == 1:
        return out[0]

    return out

def wrap_function(fcn,argv=None,optional_arg_types={}):
    """
    A generalized main function that constructs a command line argument parser
    and then parses arguments for any function.

    fcn: function to run.
    argv: arguments to parse. if none, use sys.argv[1:]
    optional_arg_types: dictionary of arg types for arguments with None as
                        their default in the function. If an argument is
                        not in optional_arg_types, the default is to treat
                        argument as an int.
    """

    # Get command line arguments
    if argv is None:
        argv = sys.argv[1:]

    # Build parser
    description = dict(inspect.getmembers(fcn))["__doc__"]
    parser = argparse.ArgumentParser(prog=f"{fcn.__name__}.py",
                                     description=description,
                                     formatter_class=argparse.RawTextHelpFormatter)

    # Build parser arguments using signature of fcn
    param = inspect.signature(fcn).parameters
    for p in param:

        # If no default specified, make required
        if param[p].default is param[p].empty:
            parser.add_argument(p)

        # If default specified, make optional
        else:

            # For type == None args, parse as string
            if param[p].default is None:
                try:
                    arg_type = optional_arg_types[p]
                except KeyError:
                    arg_type = str

            else:
                arg_type = type(param[p].default)

            parser.add_argument(f"--{p}",
                                type=arg_type,
                                default=param[p].default)

    # Parse stats
    args = parser.parse_args(argv)

    # Call function with kwargs
    fcn(**args.__dict__)



"""
Run raxml. Creates a working directory, copies in the relevant files, runs
there, and then returns to the previous directory.
"""

# raxml binary to use it not specified by user
RAXML_BINARY = "raxml-ng"

import topiary
import topiary.external._interface as interface

import pandas as pd
import multiprocessing as mp
import os

def run_raxml(algorithm=None,
              alignment_file=None,
              tree_file=None,
              model=None,
              dir_name=None,
              seed=None,
              threads=-1,
              raxml_binary=RAXML_BINARY,
              log_to_stdout=True,
              other_args=[]):
    """
    Run raxml. Creates a working directory, copies in the relevant files, runs
    there, and then returns to the previous directory.

    Parameters
    ----------
    algorithm : str
        algorithm to run (--all, --ancestral, etc.)
    alignment_file : str
        alignment file in .phy format (passed via --msa)
    tree_file : str
        tree file in .newick format (passed via --tree)
    model : str
        model in format recognized by --model
    dir_name : str,optional
        If specified, this will be the name of the working directory.
    seed : bool,int,str
        If true, pass a randomly generated seed to raxml. If int or str, use
        that as the seed. (passed via --seed)
    threads : int, default=-1
        number of threads (passed via --threads). if -1, use all available.
    raxml_binary : str, default=RAXML_BINARY
        raxml binary to use
    log_to_stdout : book, default=True
        capture log and write to std out.
    other_args : list-like
        list of arguments to pass to raxml

    Return
    ------
    raxml_command : string
        command passed to raxml-ng as a string
    """

    # Create directory in which to do calculation
    dir_name = interface.create_new_dir(dir_name=dir_name)

    # Copy alignment and tree files into the directory (if specified)
    if alignment_file is not None:
        alignment_file = interface.copy_input_file(alignment_file,
                                                   dir_name,
                                                   file_name="alignment.phy",
                                                   put_in_input_dir=False)
    if tree_file is not None:
        tree_file = interface.copy_input_file(tree_file,
                                              dir_name,
                                              file_name="tree.newick",
                                              put_in_input_dir=False)

    # Build a command list
    cmd = [raxml_binary]

    if algorithm is not None:
        cmd.append(algorithm)

    if alignment_file is not None:
        cmd.extend(["--msa",alignment_file])

    if tree_file is not None:
        cmd.extend(["--tree",tree_file])

    if model is not None:
        cmd.extend(["--model",model])

    # seed argument is overloaded. Interpret based on type
    if seed is not None:
        if type(seed) is bool:
            if seed:
                cmd.extend(["--seed",interface.gen_seed()])
        elif type(seed) is int:
            cmd.extend(["--seed",f"{seed:d}"])
        elif type(seed) is str:

            try:
                int(seed)
            except ValueError:
                err = f"seed {seed} could not be interpreted as an int\n"
                raise ValueError(err)

            cmd.extend(["--seed",seed])
        else:
            err = "seed must be True/False, int, or string representation of int\n"
            raise ValueError(err)

    # Figure out number of threads to use
    if threads < 0:
        try:
            threads = mp.cpu_count()
        except NotImplementedError:
            threads = os.cpu_count()
            if threads is None:
                print("Could not determine number of cpus. Using single thread.\n")
                threads = 1

    cmd.extend(["--threads",f"{threads:d}"])

    # Put on any custom args
    for a in other_args:
        cmd.append(a)

    # If logging to standard out, get log file name
    log_file = None
    if log_to_stdout:
        log_file = "alignment.phy.raxml.log"

    # Run job
    interface.launch(cmd,dir_name,log_file)

    return " ".join(cmd)



"""
Run BLAST against a remote NCBI database.
"""

import topiary
from topiary import check
from .util import read_blast_xml, _standard_blast_args_checker

from Bio import Entrez
from Bio.Blast import NCBIXML, NCBIWWW

import numpy as np
import pandas as pd

from tqdm.auto import tqdm

import sys, urllib, http, copy, os, time, random, string
import multiprocessing as mp

def _prepare_for_blast(sequence,
                       db,
                       taxid,
                       blast_program,
                       hitlist_size,
                       e_value_cutoff,
                       gapcosts,
                       url_base,
                       kwargs):
    """
    Take inputs to ncbi_blast, check arguments, and do initial processing.
    Outputs are a list of validated sequences, a dictionary of keyword arguments
    to past to qblast for each query, and whether or not to return a single df
    or list of df at the end.

    Parameters
    ----------
        sequence: sequence as a string OR list of string sequences
        db: NCBI blast database
        taxid: taxid for limiting blast search (default to no limit)
        blast_program: NCBI blast program to use (blastp, tblastn, etc.)
        hitlist_size: download only the top hitlist_size hits
        e_value_cutoff: only take hits with e_value better than e_value_cutoff
        gapcost: BLAST gapcosts (length 2 tuple of ints)
        url_base: NCBI base url
        kwargs: extra keyword arguments are passed directly to Bio.Blast.NCBIWWW.qblast,
                overriding anything constructed above. You could, for example, pass
                entrez_query="txid9606[ORGN] or txid10090[ORGN]" to limit blast
                results to hits from human or mouse. This would take precedence over
                any taxid specified above.

    Return
    ------
        sequence_list, blast_kwargs, return_singleton
    """

    # Deal with standard input
    out = _standard_blast_args_checker(sequence,
                                       hitlist_size,
                                       e_value_cutoff,
                                       gapcosts)
    sequence_list = out[0]
    hitlist_size = out[1]
    e_value_cutoff = out[2]
    gapcosts = out[3]
    return_singleton = out[4]

    # -------------------------------------------------------------------------
    # Deal with db input

    db_is_bad = True
    if type(db) is str:
        if len(db) > 1:
            db_is_bad = False

    if db_is_bad:
        err = "\ndb should be a string indicating an NCBI database (e.g. 'nr')\n\n"
        raise ValueError(err)

    # -------------------------------------------------------------------------
    # Deal with taxid input

    taxid_out = []
    taxid_is_bad = True

    # Is it int-like? Convert integer input to list of one string. We're
    # stringent on this -- no floats -- because an int cast will silently
    # round down. If someone used a taxid like 960.6 (extra .) it would be
    # interpreted as 960, which is very much the wrong taxid
    if np.issubdtype(type(taxid),np.integer):
        taxid = [f"{taxid}"]

    # This wacky line sees if the taxid is iterable, but not a type *class*.
    # Catches weird edge case where user passes in str or int as a taxid
    if hasattr(taxid,"__iter__") and type(taxid) is not type:

        taxid_is_bad = False

        # If taxid is a single string, convert it to a list of one string
        if type(taxid) is str:
            taxid = [taxid]

        # Go through list of taxids and put in correct format
        for t in taxid:
            if type(t) is str:
                taxid_out.append(f"txid{t}[ORGN]")
            elif np.issubdtype(type(t),np.integer):
                taxid_out.append(f"txid{t}[ORGN]")
            else:
                taxid_is_bad = True
                break

    # If taxid was None to begin with, ignore it
    else:
        if taxid is None:
            taxid_out = []
            taxid_is_bad = False

    if taxid_is_bad:
        err = "\ntaxid should be either a single ncbi taxon id (e.g. 9606 for\n"
        err += "human) or list of ncbi taxon ids. Individual ids can either be\n"
        err += "int or str. These are passed to NCBI without further validation.\n\n"
        raise ValueError(err)

    # -------------------------------------------------------------------------
    # Deal with blast_program

    blast_program_is_bad = True
    if type(blast_program) is str:
        if len(blast_program) > 1:
            blast_program_is_bad = False

    if blast_program_is_bad:
        err = "\nblast_program should be a string identifier for the blast\n"
        err += "program to use (i.e. blastp, tblastn, etc.)\n"
        raise ValueError(err)

    # -------------------------------------------------------------------------
    # Deal with url_base

    url_base_is_bad = True
    if type(url_base) is str:
        if len(url_base) > 1:
            url_base_is_bad = False

    if url_base_is_bad:
        err = "\nurl_base should be a string holding a url pointing to the\n"
        err += "blast server you wish to use.\n"
        raise ValueError(err)


    # -------------------------------------------------------------------------
    # Construct qblast keywords

    # keyword arguments to pass to qblast *besides* sequence
    blast_kwargs = {"program":blast_program,
                    "database":db,
                    "hitlist_size":f"{hitlist_size}",
                    "expect":f"{e_value_cutoff}",
                    "gapcosts":f"{gapcosts[0]} {gapcosts[1]}",
                    "url_base":url_base}

    # Construct taxid entrez_query
    if len(taxid_out) > 0:
        blast_kwargs["entrez_query"] = " or ".join(taxid_out)

    # Capture the rest of kwargs, overwriting anything automatically made
    for k in kwargs:
        blast_kwargs[k] = kwargs[k]

    return sequence_list, blast_kwargs, return_singleton

def _construct_args(sequence_list,
                    blast_kwargs,
                    max_query_length,
                    num_tries_allowed,
                    num_threads=-1,
                    test_num_cores=None):
    """
    Construct a list of arguments to pass to each thread that will run a
    blast query.

    Parameters
    ----------
        sequence_list: list of sequences as strings
        blast_kwargs: keyword arguments to pass to blast call
        max_query_length: maximum string length accepted by the server. if the
                          query is too long, this function will break it into
                          multiple requests, each sent to ncbi.
        num_tries_allowed: try num_tries_allowed times in case of timeout
        num_threads: number of threads to use (locally). if -1 use all available.
        test_num_cores: send in a hacked number of cores for testing purposes

    Return
    ------
        list of args to pass for each calculation, number of threads
    """

    # Validate inputs that have not yet been validated.
    max_query_length = check.check_int(max_query_length,
                                                   "max_query_length",
                                                   minimum_allowed=1)

    num_tries_allowed = check.check_int(num_tries_allowed,
                                                    "num_tries_allowed",
                                                    minimum_allowed=1)

    num_threads = check.check_int(num_threads,
                                              "num_threads",
                                              minimum_allowed=-1)
    if num_threads == 0:
        err = "\nnum_threads cannot be zero. It can be -1 (use all available),\n"
        err += "or any integer > 0.\n\n"
        raise ValueError(err)


    if test_num_cores is not None:
        num_cores = test_num_cores
    else:

        # Try to figure out how many cores are available
        try:
            num_cores = mp.cpu_count()
        except NotImplementedError:
            num_cores = os.cpu_count()

        # If we can't figure it out, revert to 1. (os.cpu_count() gives None
        # if the number of cores unknown).
        if num_cores is None:
            print("Could not determine number of cpus. Using single thread.",flush=True)
            num_cores = 1

    # Set number of threads
    if num_threads == -1 or num_threads > num_cores:
        num_threads = num_cores

    # Make sure sequence_list is an array
    sequence_list = np.array(sequence_list)

    # Break sequences up into blocks
    num_sequences = len(sequence_list)
    block_size = num_sequences//num_threads

    # If more threads than sequences, block_size is 1. Only run one per thread.
    if block_size < 1:
        block_size = 1
        num_threads = num_sequences

    # Windows will hold indexes for each block
    windows = [block_size for _ in range(num_sequences//block_size)]
    remainder = num_sequences % block_size
    if remainder/block_size > 0.5:
        windows.append(remainder)
    else:
        counter = 0
        while remainder > 0:
            windows[counter] += 1
            remainder -= 1
    windows.insert(0,0)

    windows = np.cumsum(windows)

    # Blocks will allow us to tile over all sequences
    split_sequences = []
    counter = 0
    for i in range(len(windows)-1):
        split_sequences.append([])
        for seq in sequence_list[windows[i]:windows[i+1]]:

            new_sequence = f">count{counter}\n{seq}\n"

            # If you have a super, super long sequence...
            if len(new_sequence) > max_query_length:
                err = "\nat least one sequence is, by itself, greater than\n"
                err += f"the maximum server query length '{max_query_length}'.\n"
                err += "To fix, either trim this sequence, remove it from the\n"
                err += "query list, increase the max_query_length (not\n"
                err += "recommended), or use a local blast database rather\n"
                err += "than the NCBI database. Sequence is:\n\n"
                err += f"{seq}\n\n"
                raise ValueError(err)

            new_length = len(new_sequence)
            current_length = len("".join(split_sequences[-1]))
            if current_length + new_length > max_query_length:
                split_sequences.append([new_sequence])
            else:
                split_sequences[-1].append(new_sequence)

            counter += 1

    final_sequences = []
    for s in split_sequences:
        final_sequences.append("".join(s))

    all_args = []
    for i, seq in enumerate(final_sequences):
        query = copy.deepcopy(blast_kwargs)
        query["sequence"] = seq
        all_args.append([query,i,num_tries_allowed])

    return all_args, num_threads


def _thread_manager(all_args,num_threads):
    """
    Run a bunch of blast jobs in a mulithreaded fashion. Should only be called
    by ncbi_blast.

    Parameters
    ----------
        all_args: list of lists containing arguments to pass to to _thread.
                  Should not have the `queue` argument, as this will be added
                  by this function.
        num_threads: number of threads to use.

    Return
    ------
        list of hits from all blast calls when all threads complete. These will
        be ordered by `counter`, passed in via all_args.
    """

    print(f"Performing {len(all_args)} blast queries on {num_threads} threads.",
          flush=True)

    # queue will hold results from each run. Append to each entry in all_args
    manager = mp.Manager()
    queue = manager.Queue()
    lock = manager.Lock()
    for i in range(len(all_args)):
        all_args[i].append(queue)
        all_args[i].append(lock)

    with mp.Pool(num_threads) as pool:

        # Black magic. pool.imap() runs a function on elements in iterable,
        # filling threads as each job finishes. (Calls _blast_thread
        # on every args tuple in all_args). tqdm gives us a status bar.
        # By wrapping pool.imap iterator in tqdm, we get a status bar that
        # updates as each thread finishes.
        list(tqdm(pool.imap(_thread,all_args),total=len(all_args)))

    # Get results out of the queue.
    results = []
    while not queue.empty():
        results.append(queue.get())

    # Sort results
    results.sort()
    hits = [r[1] for r in results]

    return hits

def _thread(args):
    """
    Run an NCBIWWW.qblast call on a single thread, making several attempts.
    Put results in a multiprocessing queue.

    args is expanded into the following:
        this_query: qblast keyword arguments
        counter: integer, used to sort results after they all complete
        num_tries_allowed: how many times to try before giving up
        queue: multiprocessing queue in which to store results

    Parameters
    ----------
        args. a list of arguments expanded as detailed above.

    Return
    ------
        None. results put into queue.
    """

    this_query = args[0]
    counter = args[1]
    num_tries_allowed = args[2]
    queue = args[3]
    lock = args[4]

    # While we haven't run out of tries...
    tries = 0
    while tries < num_tries_allowed:

        # Try to read output.
        try:

            # NCBI limits requests to 3 per second for normal users. Use a lock
            # when launching that sleeps for 0.5 seconds. This means we'll make
            # a maximum of two requests per second across all threads and
            # should avoid the NCBI limit.
            lock.acquire()
            try:
                time.sleep(0.5)
            finally:
                lock.release()

            result = NCBIWWW.qblast(**this_query)

            # Write output to an xml file. NCBI can spit out trashed XML
            # with CREATE_VIEW\n\n\n randomly injected between <hit> entries.
            # biopython chokes on the input. To fix this, pull down the XML,
            # clean up, write to a file, then pass the file handle back to
            # biopython.

            # Get rid of nastiness if present.
            contents = result.readlines()
            contents = [c for c in contents if c.strip() not in ["","CREATE_VIEW"]]

            # Write temporary file
            tmp_root = "".join([random.choice(string.ascii_letters)
                                for _ in range(10)])
            tmp_file = f"{tmp_root}_ncbi-blast-result.xml"
            f = open(tmp_file,"w")
            f.write("".join(contents))
            f.close()

            # Read output xml file and parse
            f = open(tmp_file,"r")

            # Clean up
            p = NCBIXML.parse(f)
            out = []
            for r in p:
                out.append(r)
            f.close()

            # If parsing successful, nuke temporary file
            os.remove(tmp_file)

        # If some kind of http error or timeout, set out to None
        except (urllib.error.URLError,urllib.error.HTTPError,http.client.IncompleteRead):
            out = None

        # If out is None, try again. If not, break out of loop--success!
        if out is None:
            tries += 1
            continue
        else:
            break

    # We didn't get result even after num_tries_allowed tries. Throw
    # an error.
    if out is None:
        err = "\nProblem accessing with NCBI server. We got no output after\n"
        err += f"{num_tries_allowed} attempts. This is likely due to a server\n"
        err += "timeout. Try again at a later time.\n"
        raise RuntimeError(err)

    # Parse output
    out_df = read_blast_xml(out)

    queue.put((counter,out_df))


def _combine_hits(hits,return_singleton):
    """
    Parse a list of hits resulting from a set of blast queries and return a
    list of dataframes, one for each query. Results will be ordered first by the
    order they occur in the list, then sorted by query. (Assumes queries have
    the form countNUMBER when it does sorting.

    Parameters
    ----------
        hits: list of dataframes containing hits
        return_singleton: bool. whether or not to return a single dataframe or
                          list of dataframes

    Return
    ------
        combined blast hits (list of dataframes or single dataframe)
    """

    all_df = []
    for hit in hits:

        # Break big dataframe into a list of dataframes, one for each query sequence
        queries = np.unique(hit["query"])

        # Trim off "count" from "countNUMBER" queries and use to sort
        query_order = [(q[5:],q) for q in queries]
        query_order.sort()

        # Go through all queries in this hit dataframe
        for q in query_order:

            # Get dataframe
            this_df = hit.loc[hit["query"] == q[1],:]

            # No hits, return empty dataframe
            if len(this_df) == 1 and pd.isna(this_df["accession"].iloc[0]):
                all_df.append(pd.DataFrame())

            # Record hits
            else:
                all_df.append(this_df)


    # Singleton hit -- pop out of list
    if return_singleton:
        all_df = all_df[0]

    return all_df


def ncbi_blast(sequence,
               db="nr",
               blast_program="blastp",
               taxid=None,
               hitlist_size=100,
               e_value_cutoff=0.001,
               gapcosts=(11,1),
               url_base="https://blast.ncbi.nlm.nih.gov/Blast.cgi",
               max_query_length=80000,
               num_tries_allowed=5,
               num_threads=1,
               **kwargs):
    """
    Perform a blast query against a remote NCBI blast database. Takes a sequence
    or list of sequences and returns a list of topiary dataframes containing hits
    for each sequence.

    Parameters
    ----------
    sequence : str or list
        sequence as a string OR list of string sequences
    db : str, default="nr"
        name of ncbi blast database
    blast_program : str, default="blastp"
        NCBI blast program to use (i.e. "blastp", "tblastn", etc.)
    taxid : str or int or list or None, default=None
        limit ncbi blast search to specified taxid. If None, do not limit
        search. If single str/int, interpret as an integer taxid (i.e. 9606 for
        Homo sapiens). If list of str/int, interpret as multiple taxid (i.e.
        [9606,10090] for Homo sapiens OR Mus musculus).
    hitlist_size : int, default=100
        return only the top hitlist_size hits
    e_value_cutoff : float, default=0.001
        only return hits with e_value better than e_value_cutoff
    gapcosts : tuple, default=(11,1)
        BLAST gapcosts (length 2 tuple of ints)
    url_base : str, default="https://blast.ncbi.nlm.nih.gov/Blast.cgi"
        NCBI base url
    max_query_length : int, default=80000
        maximum string length accepted by the server. if the query is longer
        than max_query_length, the function will break it into multiple requests
        shorter than max_query_length, then combine the results.
    num_tries_allowed : int, default=5
        try num_tries_allowed times in case of server timeout
    num_threads : int, default=1
        number of threads to use. if -1, use all available.
    **kwargs : dict, optional
        extra keyword arguments are passed directly to Bio.Blast.NCBIWWW.qblast,
        overriding anything constructed above. You could, for example, pass
        entrez_query="txid9606[ORGN] or txid10090[ORGN]" to limit blast
        results to hits from human or mouse. This would take precedence over
        any taxid specified above.

    Returns
    -------
    blast_output : pandas.DataFrame or list
        If a single sequence is passed in, return a dataframe of hits. If a list
        of queries is passed in, returns a list of dataframes--one entry per
        query. If no hits are found for a given query, return an empty dataframe.
    """

    # Generate a list of sequences, keywords to pass to qblast, and whether or
    # not to return a single or list of dataframes
    prep = _prepare_for_blast(sequence=sequence,
                              db=db,
                              taxid=taxid,
                              blast_program=blast_program,
                              hitlist_size=hitlist_size,
                              e_value_cutoff=e_value_cutoff,
                              gapcosts=gapcosts,
                              url_base=url_base,
                              kwargs=kwargs)

    sequence_list = prep[0]
    blast_kwargs = prep[1]
    return_singleton = prep[2]

    # Get arguments to pass to each thread
    all_args, num_threads = _construct_args(sequence_list=sequence_list,
                                            blast_kwargs=blast_kwargs,
                                            max_query_length=max_query_length,
                                            num_threads=num_threads,
                                            num_tries_allowed=num_tries_allowed)

    # Run multi-threaded blast
    hits = _thread_manager(all_args,num_threads)

    # Combine hits into dataframes, one for each query
    out_df = _combine_hits(hits,return_singleton)

    return out_df



"""
Functions and classes for plotting ete3 trees.
"""

import topiary
import ete3

import numpy as np
import re, os

class ColorMap:
    """
    Class for mapping a value to a specific RGB value, returned as a hexadecimal
    value interpretable by ete3.
    """

    def __init__(self,span=(0,1),R=(1,0),G=(1,0),B=(1,0)):
        """
        Initialize class instance.

        Parameters
        ----------
            span: tuple holding minimum and maximum values for color map.
            R: tuple holding red channel value corresponding to minimum and maximum
               values of span. Values should be between 0 and 1.
            G: tuple holding green channel value corresponding to minimum and maximum
               values of span. Values should be between 0 and 1.
            B: tuple holding blue channel value corresponding to minimum and maximum
               values of span. Values should be between 0 and 1.
        """

        # Deal with span
        try:
            if len(span) != 2:
                raise ValueError

            self._min_value = float(span[0])
            self._max_value = float(span[1])

            if self._min_value >= self._max_value:
                raise ValueError

        except (IndexError,ValueError,TypeError,KeyError):
            err = "\nspan should be a length-2 tuple holding the minimum and\n"
            err += "maximum values corresponding to the low and high values \n"
            err += "of the ColorMap. For example, if bootstrap support values\n"
            err += "range from 0-100, span should be (0,100). The second value\n"
            err += "must be larger than the first value.\n"
            raise ValueError(err)

        # Deal with red channel
        try:
            if len(R) != 2:
                raise ValueError

            self._R_bottom = float(R[0])
            self._R_top = float(R[1])

            if self._R_bottom < 0 or self._R_bottom > 1:
                raise ValueError

            if self._R_top < 0 or self._R_top > 1:
                raise ValueError

        except (IndexError,ValueError,TypeError,KeyError):
            err = "\nR should be a length-2 tuple holding the minimum and\n"
            err += "maximum red channel values that correspond to the low and\n"
            err += "high values of this channel in the ColorMap. For example,\n"
            err += "if the channel should go from off to on over the map, this\n"
            err += "should be (0,1). These values must be between 0 and 1.\n"
            raise ValueError(err)

        # Determine slope and intercept for R channel
        self._R_slope = (self._R_top - self._R_bottom)
        self._R_intercept = self._R_bottom

        # Deal with green channel
        try:
            if len(G) != 2:
                raise ValueError

            self._G_bottom = float(G[0])
            self._G_top = float(G[1])

            if self._G_bottom < 0 or self._G_bottom > 1:
                raise ValueError

            if self._G_top < 0 or self._G_top > 1:
                raise ValueError

        except (IndexError,ValueError,TypeError,KeyError):
            err = "\nG should be a length-2 tuple holding the minimum and\n"
            err += "maximum green channel values that correspond to the low and\n"
            err += "high values of this channel in the ColorMap. For example,\n"
            err += "if the channel should go from off to on over the map, this\n"
            err += "should be (0,1). These values must be between 0 and 1.\n"
            raise ValueError(err)

        # Determine slope and intercept for R channel
        self._G_slope = (self._G_top - self._G_bottom)
        self._G_intercept = self._G_bottom

        # Deal with green channel
        try:
            if len(B) != 2:
                raise ValueError

            self._B_bottom = float(B[0])
            self._B_top = float(B[1])

            if self._B_bottom < 0 or self._B_bottom > 1:
                raise ValueError

            if self._B_top < 0 or self._B_top > 1:
                raise ValueError

        except (IndexError,ValueError,TypeError,KeyError):
            err = "\nGBshould be a length-2 tuple holding the minimum and\n"
            err += "maximum blue channel values that correspond to the low and\n"
            err += "high values of this channel in the ColorMap. For example,\n"
            err += "if the channel should go from off to on over the map, this\n"
            err += "should be (0,1). These values must be between 0 and 1.\n"
            raise ValueError(err)

        # Determine slope and intercept for R channel
        self._B_slope = (self._B_top - self._B_bottom)
        self._B_intercept = self._B_bottom


    def hex(self,value):
        """
        Return the hexadecimal representation of the RGB value corresponding to
        "value" given the span and R,G,B defined when the class was initialized.
        """

        try:
            value = float(value)
        except (ValueError,TypeError):
            err = "\nvalue must be a float\n\n"
            raise ValueError(err)

        # Force value to be within specified min/max range
        if value < self._min_value:
            value = self._min_value
        if value > self._max_value:
            value = self._max_value

        # Scale value to between 0 and 1.
        scaled = (value - self._min_value)/(self._max_value - self._min_value)

        # Determine RGB for this value
        R = int(round(255*(self._R_intercept + self._R_slope*scaled),0))
        G = int(round(255*(self._G_intercept + self._G_slope*scaled),0))
        B = int(round(255*(self._B_intercept + self._B_slope*scaled),0))

        # Return as a string HEX value
        return f"#{bytearray((R,G,B)).hex()}"

def load_trees(prev_run_dir,
               tree_base_path,
               tree_names,
               tree_fmt=None,
               outgroup=None):
    """
    Load trees and tree formatting information from a previous run directory.

    Parameters
    ----------
        prev_run_dir: previous run directory (for example, output from an ML tree
                      inference)
        tree_base_path: base directory to look for trees within prev_run_dir. (for
                        example, look in ml_tree/output for output trees for
                        ML tree inference)
        tree_names: list of tree files to look for in tree_base_path
        tree_fmt: ete3 Tree format to use for reading trees. Should be integers.
                  List, indiciating which tree format should be applied to read each
                  file in tree_names. Should be the same length as tree_names.
        df: topiary dataframe to use to interpret tip_columns. overrides the
            dataframe that is pulled from prev_run_dir.

    Return
    ------
        ete3 Tree instances for each tree requested in tree_names.
    """

    # If the user did not specify tree formats, make them all 1
    if tree_fmt is None:
        tree_fmt = [1 for _ in tree_names]

    # Check tree format and tree names compatibility
    if len(tree_names) != len(tree_fmt):
        err = "\ntree_fmt length does not match tree_names length\n\n"
        raise ValueError(err)

    # Go over all tree names
    T_list = []
    for i, t in enumerate(tree_names):

        tree_path = os.path.join(prev_run_dir,tree_base_path,t)
        if not os.path.exists(tree_path):
            err = f"\nCould not find tree file '{tree_path}'\n\n"
            raise FileNotFoundError(err)

        # read tree
        T = topiary.io.read_tree(tree_path,fmt=tree_fmt[i])

        # If outgroup specified
        if outgroup is not None:

            # Root tree. We have to try with both the left and right
            # descendants of the previous root.
            try:
                common_anc = T.get_common_ancestor(outgroup[0])
                T.set_outgroup(common_anc)
            except ete3.coretype.tree.TreeError:
                common_anc = T.get_common_ancestor(outgroup[1])
                T.set_outgroup(common_anc)

        T_list.append(T)

    if len(T_list) == 1:
        return T_list[0]

    return T_list

def create_name_dict(df,tip_columns=None,separator="|"):
    """
    Create a dictionary mapping between uid and pretty names extracted from
    tip_columns in the dataframe.

    Parameters
    ----------
        df: topiary dataframe
        tip_columns: columns in dataframe to use to create human-readable names for
                     tips on tree. For example, tip_columns = ["species","nickname"]
                     would give names like "Homo sapiens|LY96". If None, try
                     ["species","nickname"]. If nickname is not in dataframe,
                     fall back to ["species",f"{name[:10]}..."]
        separator: separator to use between pretty names

    Return
    ------
        name_dict
    """

    local_df = df
    if tip_columns is None:

        try:
            local_df.nickname
            tip_columns = ["species","nickname"]
        except AttributeError:

            local_df = df.copy()
            trunc_name = []
            for i in range(len(local_df)):
                if len(local_df["name"].iloc[i]) > 10:
                    trunc_name.append(f"{local_df['name'].iloc[i][:10]}...")
                else:
                    trunc_name.append(f"{local_df['name'].iloc[i]}")

            local_df["trunc_name"] = trunc_name
            tip_columns = ["species","trunc_name"]

    # Make sure the tip columns are sane
    for c in tip_columns:
        try:
            local_df[c]
        except KeyError:
            err = f"\ncolumn '{c}' not in dataframe. Please check the\n"
            err += "tip_columns argument.\n\n"
            raise ValueError(err)

    # Construct name_dict, which will map between the uid on the leaves to a
    # prettier/more useful name
    name_dict = {}
    for i in range(len(local_df)):
        key = local_df.uid.iloc[i]

        v = []
        for column in tip_columns:
            v.append(f"{local_df[column].iloc[i]}")
        value = separator.join(v)

        name_dict[key] = value

    return name_dict


def setup_generic_tree_format(num_leaves,
                              value_range=(0,100),
                              R=(1,0),G=(1,0),B=(1,1),
                              width=250,space_between_taxa=10,line_width=2):
    """
    Set up formating for a generic tree.

    Parameters
    ----------
        num_leaves: number of leaves in the tree
        value_range: value range for color map. if None, do not create a ColorMap.
        R: tuple holding red channel value corresponding to minimum and maximum
           values of span. Values should be between 0 and 1.
        G: tuple holding green channel value corresponding to minimum and maximum
           values of span. Values should be between 0 and 1.
        B: tuple holding blue channel value corresponding to minimum and maximum
           values of span. Values should be between 0 and 1.
        width: width of total tree (pixels)
        height: height of total tree (pixels)
        line_width: width of tree lines (pixels)

    Returns
    -------
        cm (color map)
        ts: ete3.TreeStyle object
        ns: ete3.NodeStyle object
    """

    # Decide whether/how we're doing color map
    if value_range is not None:
        cm = ColorMap(span=value_range,R=R,G=G,B=B)
    else:
        cm = None

    # Construct basic tree format
    ts = ete3.TreeStyle()
    ts.show_scale = True
    ts.scale = width
    ts.branch_vertical_margin = space_between_taxa

    # Construct basic node format
    ns = ete3.NodeStyle()
    ns["size"] = 0
    ns["hz_line_width"] = line_width
    ns["vt_line_width"] = line_width

    return cm, ts, ns

def final_render(T,ts,output_file,default_pdf):
    """
    Render tree T in a sane fashion.

    Parameters
    ----------
        T: ete3.Tree object to draw
        ts: ete3.TreeStyle object to use to format tree object
        output_file: write tree to output file. If None and in a notebook,
                     just draw on the screen. If None and *not* in a notebook,
                     write only to default_pdf
        default_pdf: write to this pdf file. if None, do not write.

    Returns
    -------
        T.render output (if in a notebook) or None (if not in a notebook. )
    """

    # If output_file is not specified...
    if output_file is None:

        # If not in a notebook, create a pdf
        if not topiary._in_notebook:
            output_file = default_pdf

    # If output_file specified, write out
    if output_file is not None:
        T.render(output_file,tree_style=ts)

    # If this is in a notebook, render it and return so it appears
    if topiary._in_notebook:
        return T.render("%%inline",tree_style=ts)

    return None



"""
Interface to NCBI blast.
"""

from . util import read_blast_xml, _standard_blast_args_checker

from . ncbi import ncbi_blast
from . local import local_blast
from . recip import recip_blast
from . merge import merge_blast_df
from . make import make_blast_db



"""
Reciprocal blast sequences.
"""

import topiary
from topiary import check
from .ncbi import ncbi_blast
from .local import local_blast

from topiary.external.ncbi import parse_ncbi_line

import pandas as pd
import numpy as np

import re, sys, copy

def _prepare_for_blast(df,
                       paralog_patterns,
                       local_blast_db,
                       ncbi_blast_db,
                       ignorecase,
                       max_del_best,
                       min_call_prob,
                       use_start_end):
    """
    Check sanity of input parameters and compile patterns. Return a validated
    topiary dataframe, list of sequences, compiled set of patterns to search,
    and sanitized parameters.

    Parameters
    ----------
        df: topiary dataframe. Will pull sequences from df.sequences. If there are
            'start' and 'stop' columns in the dataframe, only blast sequences
            between start/top (for example: start = 5, stop = 20 would blast
            sequence[5:20]. To turn this off, set use_start_end = False.
        paralog_patterns: dictionary with paralogs as values and lists of patterns
                          to look for as values.
        local_blast_db: local database against which to blast
        ncbi_blast_db: database on ncbi against which to blast
        ignorecase: whether to ignore letter case when searching blast (default True).
        max_del_best: maximum value for log(e_hit_match) - log(e_hit_best) that
                      allows for paralog call. This means the matched recip blast
                      hit does not have to be the best hit, but must be within this
                      e-value difference of the best hit. A higher number means a
                      less stringent cutoff. A value of 0 would require the matched
                      hit be the best hit.
        min_call_prob: hits from all paralogs that yield a regular expression match
                       are weighted by their relative e-values. Each paralog is
                       assigned a relative probability. This cutoff is the minimum
                       probability the best paralog match must have to result in a
                       paralog call. Value should be between 0 and 1 (not inclusive),
                       where min_call_prob --> 1 increases the stringency.
        use_start_end: boolean. whether or not to use start/stop columns in
                        dataframe (if present) to slice subset of sequence to blast.

    Return
    ------
        df, sequence_list, patterns, max_del_best, min_call_prob
    """

    # Check type of df
    if type(df) is not pd.DataFrame:
        err = "\ndf should be a topiary dataframe\n\n"
        raise ValueError(err)

    # Check ignorecase
    try:
        ignorecase = bool(int(ignorecase))
    except (ValueError,TypeError):
        err = "\nignorecase must be True or False\n\n"
        raise ValueError(err)

    patterns = check.check_paralog_patterns(paralog_patterns,
                                                        ignorecase=ignorecase)
    if len(patterns) == 0:
        err = "\nparalog_patterns must have at least one entry\n"
        raise ValueError(err)

    # Validate blast database arguments
    if ncbi_blast_db is None and local_blast_db is None:
        err = "\nPlease specificy either ncbi_blast_db OR local_blast_db\n\n"
        raise ValueError(err)

    if ncbi_blast_db is not None and local_blast_db is not None:
        err = "\nPlease specificy either ncbi_blast_db OR\n"
        err += "local_blast_db, but not both.\n\n"
        raise ValueError(err)

    # Check max_del_best
    try:
        max_del_best = float(max_del_best)
        if max_del_best < 0 or np.isnan(max_del_best):
            raise ValueError

    except (ValueError,TypeError):
        err = "\nmax_del_best must be a number between 0 and positive infinity\n\n"
        raise ValueError(err)

    # Check min_call_prob
    try:
        min_call_prob = float(min_call_prob)
        if min_call_prob <= 0 or min_call_prob >= 1:
            raise ValueError
        if np.isnan(min_call_prob):
            raise ValueError
    except (ValueError,TypeError):
        err = "\nmin_call_prob must be a number between 0 and 1 (not inclusive)\n\n"
        raise ValueError(err)

    # Check use_start_end
    try:
        use_start_end = bool(int(use_start_end))
    except (ValueError,TypeError):
        err = "\nuse_start_end must be True or False\n\n"
        raise ValueError(err)

    # Make sure dataframe is a topiary dataframe
    df = check.check_topiary_dataframe(df)

    # Create list of all sequences in dataframe
    sequence_list = []
    for idx in df.index:

        # Do not recip blast sequences where we already have keep == False
        if not df.loc[idx,"keep"]:
            continue

        # Get sequence
        s = df.loc[idx,"sequence"]

        # Try to get start/end (only blasts subset of sequences). If start and
        # end are not defined in this dataframe, take whole sequence
        try:
            a = df.loc[idx,"start"]
            b = df.loc[idx,"end"]
        except KeyError:
            a = 0
            b = None

        # Ignore start/stop if specified.
        if not use_start_end:
            a = 0
            b = None

        # Record this sequence
        try:
            sequence_list.append(s[a:b])
        except TypeError:
            err = "\nstart and end columns must be integers\n\n"
            raise ValueError(err)

        # Warn if the sequence is empty after slicing
        if sequence_list[-1] == "":
            err = "\nsequence has zero length (after slicing with start/end).\n"
            err += f"row: {df.loc[idx,:]}\n\n"
            raise ValueError(err)

    return df, sequence_list, patterns, max_del_best, min_call_prob


def _run_blast(sequence_list,
               local_blast_db,
               ncbi_blast_db,
               ncbi_taxid,
               hitlist_size,
               e_value_cutoff,
               gapcosts,
               num_threads,
               **kwargs):
    """
    Run blast on sequence_sequence list, returning a list of dataframes -- one
    df of hits for each sequence in sequence_list.

    Parameters
    ----------
        sequence_list: list of sequences to use as blast queries.
        local_blast_db: local database against which to blast
        ncbi_blast_db: database on ncbi against which to blast
        ncbi_taxid: limit search to species specified by taxid for an ncbi search.
        e_value_cutoff: minimum allowable e value for a hit
        gapcosts: gap costs (must be length 2 tuple of ints)
        num_threads: number of threads to use for blast search. if -1,
                           use all available.
        kwargs: extra keyword arguments are passed directly to biopython
                NcbiblastXXXCommandline (for local blast) or qblast (for remote
                blast). These take precedence over anything specified above
                (hitlist_size, for example).

    Return
    ------
        list of dataframes with blast hits, one for each sequence in sequence_list
    """

    # NCBI blast
    if ncbi_blast_db:

        try:
            taxid = kwargs.pop("taxid")
            if ncbi_taxid is None:
                ncbi_taxid = taxid
            else:
                err = "\nplease specify the taxid to use for an ncbi blast search\n"
                err += "using the `ncbi_taxid` keyword argument.\n"
                raise ValueError(err)
        except KeyError:
            pass

        # Warn that NCBI blasting can be slow
        w = "\nBlasting against the NCBI database can be slow/unstable. Consider\n"
        w += "creating a local BLAST database for your reciprocal BLAST needs.\n"
        print(w,flush=True)

        hit_dfs = ncbi_blast(sequence_list,
                             db=ncbi_blast_db,
                             taxid=ncbi_taxid,
                             blast_program="blastp",
                             hitlist_size=hitlist_size,
                             e_value_cutoff=e_value_cutoff,
                             gapcosts=gapcosts,
                             num_threads=num_threads,
                             **kwargs)

    # Local blast
    else:
        hit_dfs = local_blast(sequence_list,
                              db=local_blast_db,
                              blast_program="blastp",
                              hitlist_size=hitlist_size,
                              e_value_cutoff=e_value_cutoff,
                              gapcosts=gapcosts,
                              num_threads=num_threads,
                              **kwargs)

    return hit_dfs


def _make_recip_blast_calls(df,
                              hit_dfs,
                              patterns,
                              max_del_best,
                              min_call_prob,
                              ncbi_blast_db):
    """
    Make paralog calls given blast output and list of patterns.

    Parameters
    ----------
        df: topiary dataframe with query sequences
        hit_dfs: list of dataframes returned by blast
        patterns: dictionary with paralogs as values and lists of patterns
                  to look for as values.
        max_del_best: maximum value for log(e_hit_match) - log(e_hit_best) that
                      allows for paralog call. This means the matched recip blast
                      hit does not have to be the best hit, but must be within this
                      e-value difference of the best hit. A higher number means a
                      less stringent cutoff. A value of 0 would require the matched
                      hit be the best hit.
        min_call_prob: hits from all paralogs that yield a regular expression match
                       are weighted by their relative e-values. Each paralog is
                       assigned a relative probability. This cutoff is the minimum
                       probability the best paralog match must have to result in a
                       paralog call. Value should be between 0 and 1 (not inclusive),
                       where min_call_prob --> 1 increases the stringency.
        ncbi_blast_db: database used for ncbi blast. (Used to determine if
                           this should be parsed as ncbi vs. local blast inputs)

    Return
    ------
        dataframe with recip blast calls
    """

    # hit_dfs is a list of dataframes, each of which has the blast hits for
    # each sequence in the input topiary dataframe.

    # Go through hits from each sequence
    results = {"recip_found_paralog":[],
               "recip_hit":[],
               "recip_paralog":[],
               "recip_prob_match":[],
               "recip_del_best":[]}

    for _, hits in enumerate(hit_dfs):

        # No recip blast hits at all for this sequence
        if len(hits) == 0:
            results["recip_found_paralog"].append(False)
            results["recip_hit"].append(pd.NA)
            results["recip_paralog"].append(pd.NA)
            results["recip_prob_match"].append(np.nan)
            results["recip_del_best"].append(np.nan)
            continue

        # Get e value of top hit
        top_e_value = hits["e_value"].iloc[0]

        # Now go through each regular expression pattern
        e_values = []
        hit_defs = []
        paralogs = []
        for i, p in enumerate(patterns):

            # Go through each hit description...
            for j, description in enumerate(hits.hit_def):

                row = hits.iloc[j]

                # If the pattern matches this hit description
                if p[0].search(description):

                    # Get hit definition
                    this_def = row["hit_def"]

                    # If this was an NCBI blast, try to parse the NCBI line,
                    # pulling apart multi-titles
                    if ncbi_blast_db:
                        hd = parse_ncbi_line(row["title"],row["accession"])
                        if hd is not None:
                            this_def = hd["name"]

                    hit_defs.append(this_def)

                    # get e value
                    e_values.append(row["e_value"])

                    # Get paralog call
                    paralogs.append(p[1])

                    break


        # If we got at least one hit that matched
        recip_found_paralog = False
        if len(paralogs) > 0:

            # calculate posterior probability
            e_values = np.array(e_values)

            # If none of the e-values are zero, calculate the posterior
            # probability of the top hit.
            if np.sum(e_values == 0) == 0:
                pp = 1/np.array(e_values)
                pp = pp/np.sum(pp)
                match_idx = np.argsort(pp)[-1]
                match_pp = pp[match_idx]

            # If there is a zero e-value, take the first hit a zero e-value.
            # This is an (apparently) infinite posterior probaility
            else:
                match_idx = np.where(e_values == 0)[0][0]
                match_pp = np.inf

            # Get the match hit definition, paralog call, match probability
            # (relative to other possible paralogs), and difference in e value
            # relative to the best overall hit.
            recip_hit = hit_defs[match_idx]
            recip_paralog = paralogs[match_idx]
            recip_prob_match = match_pp
            recip_e_value = e_values[match_idx]

            # If the top hit has e-value of zero...
            if top_e_value == 0:

                # If the recip blast hit matches this, the difference in those
                # e-values is zero
                if recip_e_value == 0:
                    recip_del_best = 0

                # If the recip blast hit does *not* match the top, the
                # difference in those e-values is
                else:
                    recip_del_best = -np.inf
                    w = "WARNING:\n"
                    w += "BLAST returned an e-value = 0.0 for the top hit:\n\n"
                    w += f"  {hits['hit_def'].iloc[0]}\n\n"
                    w += "This hit does not match any of the patterns in paralog_patterns.\n"
                    w += "The best hit that matches something in paralog_patterns has\n"
                    w += f"an e-value of {recip_e_value} and matches '{recip_paralog}'.\n"
                    w += "These values cannot be quantitatively compared.\n"
                    w += "Topiary will set recip_found_paralog to 'False' for\n"
                    w += "this sequence.\n"
                    print(w)

            # If we get here, neither top_e_value or recip_e_value are zero.
            # (if recip_e_value was zero, it would be top_e_value and thus be
            # caught above.)
            else:
                recip_del_best = np.log10(top_e_value) - np.log10(recip_e_value)

            # Decide if we can make a paralog call based on del_best and
            # prob_match
            if recip_del_best < max_del_best and recip_prob_match > min_call_prob:
                recip_found_paralog = True

        else:
            recip_hit = hits["hit_def"].iloc[0]
            recip_paralog = None
            recip_prob_match = None
            recip_del_best = None

        # Record results of analysis
        results["recip_found_paralog"].append(recip_found_paralog)
        results["recip_hit"].append(recip_hit)
        results["recip_paralog"].append(recip_paralog)
        results["recip_prob_match"].append(recip_prob_match)
        results["recip_del_best"].append(recip_del_best)

    # Only load results for values with keep == True
    for k in results:

        # If recip_found_paralog, set to False by default. Overwritten with
        # Trues below
        if k == "recip_found_paralog":
            df[k] = False

        # Otherwise, set to a missing value.
        else:
            df[k] = pd.NA

        # Now set keep with the results from this column
        df.loc[df.keep,k] = results[k]

    num_keep_at_start = np.sum(df.keep)

    # Update keep with recip_found_paralog
    df.loc[:,"keep"] = np.logical_and(df["keep"],df["recip_found_paralog"])

    # Write out some summary statistics
    print(f"{np.sum(df.keep)} of {num_keep_at_start} sequences passed recip blast.\n")
    print("Found the following numbers of paralogs:")
    for p in patterns:
        num_of_p = np.sum(df.recip_paralog == p[1])
        print(f"    {p[1]}: {num_of_p}")
    print("",flush=True)

    return df


def recip_blast(df,
                  paralog_patterns,
                  local_blast_db=None,
                  ncbi_blast_db=None,
                  ncbi_taxid=None,
                  ignorecase=True,
                  max_del_best=100,
                  min_call_prob=0.95,
                  use_start_end=True,
                  hitlist_size=50,
                  e_value_cutoff=0.01,
                  gapcosts=(11,1),
                  num_threads=-1,
                  **kwargs):
    """
    Take sequences from a topiary dataframe and do a recip blast analysis
    against an NCBI or local blast database. Looks in blast hits for the
    regular expressions defined in paralog_patterns to call paralog for each
    sequence in df. Returns a copy of the input topiary dataframe with five new
    columns:

    + `recip_found_paralog`: True/False, whether a paralog was found
    + `recip_hit`: string, description for paralog hit (if found) or best hit
      (if no match found)
    + `recip_paralog`: string or None. name of paralog from paralog_patterns
    + `recip_prob_match`: float. probability that this is the correct paralog
      call based on relative evalues of all paralog hits
    + `recip_del_best`: float. how much worse paralog call is than the best
      blast hit. log(e_hit_match) - log(e_hit_best)

    Parameters
    ----------

    df : pandas.DataFrame
        topiary dataframe. Will pull sequences from df.sequences. If there are
        'start' and 'stop' columns in the dataframe, only blast sequences
        between start/top (for example: start = 5, stop = 20 would blast
        sequence[5:20]. To turn this off, set use_start_end = False.
    paralog_patterns : dict
        dictionary with paralogs as values and lists of patterns to look for as
        values. Keys must be strings. Values may be strings or compiled
        re.Pattern instances. See Notes for details.
    local_blast_db : str or None, default=None
        Local blast database to use. If None, use an NCBI database. Incompatible
        with ncbi_blast_db.
    ncbi_blast_db : str or None, default=None
        NCBI blast database to use. If None, use a local database. Incompatible
        with local_blast_db.
    ncbi_taxid : str or int or list or None, default=None
        limit ncbi blast search to specified taxid. If None, do not limit
        search. If single str/int, interpret as an integer taxid (i.e. 9606 for
        Homo sapiens). If list of str/int, interpret as multiple taxid (i.e.
        [9606,10090] for Homo sapiens OR Mus musculus).
    ignorecase : bool, default=True
        whether to ignore letter case when searching blast. NOTE: if you pass
        in compiled regular expressions, the flags (such as re.IGNORECASE) on
        the *last* pattern for each paralog will be used for all patterns.
    max_del_best : float, default=100
        maximum value for log(e_hit_match) - log(e_hit_best) that
        allows for paralog call. This means the matched recip blast
        hit does not have to be the best hit, but must be within this
        e-value difference of the best hit. A higher number means a
        less stringent cutoff. A value of 0 would require the matched
        hit be the best hit.
    min_call_prob: float, default=0.95
        hits from all paralogs that yield a regular expression match
        are weighted by their relative e-values. Each paralog is
        assigned a relative probability. This cutoff is the minimum
        probability the best paralog match must have to result in a
        paralog call. Value should be between 0 and 1 (not inclusive),
        where min_call_prob --> 1 increases the stringency.
    use_start_end : bool, default=True
        whether or not to use start/stop columns in dataframe (if present) to
        slice subset of sequence for reciprocal blast.
    hitlist_size : int, default=100
        return only the top hitlist_size hits
    e_value_cutoff : float, default=0.001
        only return hits with e_value better than e_value_cutoff
    gapcosts : tuple, default=(11,1)
        BLAST gapcosts (length 2 tuple of ints)
    num_threads : int, default=-1
        number of threads to use. if -1, use all available. (Multithreading
        rarely speeds up remote BLAST).
    **kwargs : dict, optional
        extra keyword arguments are passed directly to biopython
        blast). These take precedence over anything specified above
        (hitlist_size, for example).

    Returns
    ------
    topiary_dataframe : pandas.DataFrame
        copy of input dataframe with new recip blast columns

    Notes
    -----
    :code:`paralog_patterns` are used to match reciprocal blast hits. For
    example:

    .. code-block:: python

        paralog_patterns = {"LY96":["lymphocyte antigen 96","MD-2"],
                            "LY86":["lymphocyte antigen 86","MD-1"]}


    This would mean hits with 'lymphocyte antigen 96' or 'MD-2' will map to
    LY96; hits with 'lymphocyte antigen 86' or 'MD-1' will map to LY86. Hits
    that match neither would not be assigned a paralog.

    NOTE: string patterns are interpreted literally. The pattern :code:`"[A-Z]"`
    would be escaped to look for :code:`"\[A\-Z\]"`. If you want to use regular
    expressions in your patterns, pass them in as compiled regular expressions.
    For example,

    .. code-block:: python

        my_pattern = re.compile("[A-Z]")


    """

    # Check sanity of input parameters and return a validated topiary dataframe,
    # list of sequences, and compiled set of patterns to search. Note: the
    # values of ncbi_blast_db, local_blast_db, ncbi_taxid, hitlist_size,
    # gapcosts, num_threads, and kwargs are checked by the blast functions
    # themselves.
    out = _prepare_for_blast(df,
                             paralog_patterns,
                             local_blast_db,
                             ncbi_blast_db,
                             ignorecase,
                             max_del_best,
                             min_call_prob,
                             use_start_end)

    df, sequence_list, patterns, max_del_best, min_call_prob = out


    # Run BLAST on sequence list, returning list of dataframes -- one for each
    # seqeunce in sequence_list
    hit_dfs = _run_blast(sequence_list,
                         local_blast_db,
                         ncbi_blast_db,
                         ncbi_taxid,
                         hitlist_size,
                         e_value_cutoff,
                         gapcosts,
                         num_threads,
                         **kwargs)

    # Make paralog calls given blast output and list of patterns
    out_df = _make_recip_blast_calls(df,
                                       hit_dfs,
                                       patterns,
                                       max_del_best,
                                       min_call_prob,
                                       ncbi_blast_db)

    return out_df



"""
Use entrez to download protein sequences from the NCBI.
"""

import topiary
from topiary import check

from Bio import Entrez

from tqdm.auto import tqdm

import os, urllib, http, time
import multiprocessing as mp

def _get_sequences_thread(args):
    """
    Download ids from the NCBI. args is interpreted as:

    Parameters
    ----------
        args: list of arguments controlling entrez query. Expanded to:
            index: number of request, used for sorting results after multithreading
            ids: string with comma-separated list of ids to download
            num_tries_allowed: how many attempts should be made to actually do
                               a given query
            lock: lock to control number of requests per thread per second
            queue: multiprocessing queue in which to store results
    Return
    ------
        None
    """

    index = args[0]
    ids = args[1]
    num_tries_allowed = args[2]
    lock = args[3]
    queue = args[4]

    # While we haven't run out of tries...
    tries = 0
    while tries < num_tries_allowed:

        # Try to read output.
        try:

            # NCBI limits requests to 3 per second for normal users. Use a lock
            # when launching that sleeps for 0.5 seconds. This means we'll make
            # a maximum of two requests per second across all threads and
            # should avoid the NCBI limit.
            lock.acquire()
            try:
                time.sleep(0.5)
            finally:
                lock.release()

            handle = Entrez.efetch(db="protein",
                                   id=ids,
                                   rettype="fasta",
                                   retmode="text")
            out = handle.read()
            handle.close()

        # If some kind of http error or timeout, set out to None
        except (urllib.error.HTTPError,http.client.IncompleteRead):
            out = None

        # If out is None, try again. If not, break out of loop--success!
        if out is None:
            tries += 1
            continue
        else:
            break

    # We failed to download. Throw an error.
    if out is None:
        err = "could not download sequences\n"
        raise ValueError(err)

    # Record out and initial index in the output queue.
    queue.put((index,out))

def get_sequences(to_download,
                  block_size=50,
                  num_tries_allowed=10,
                  num_threads=-1):
    """
    Use entrez to download protein sequences from the NCBI.

    Parameters
    ----------
    to_download : list
        list of ncbi ids to download
    block_size : int, default=50
        download in chunks this size
    num_tries_allowed : int, default=10
        number of times to try before giving up and throwing an error.
    num_threads : int, default=-1
        number of threads to use. if -1, use all available.

    Returns
    -------
    fasta_file : str
        file holding all downloaded sequences.
    """

    block_size = check.check_int(block_size,
                                             "block_size",
                                             minimum_allowed=1)
    num_tries_allowed = check.check_int(num_tries_allowed,
                                                    "num_tries_allowed",
                                                    minimum_allowed=1)
    num_threads = check.check_int(num_threads,
                                              "num_threads",
                                              minimum_allowed=-1)
    if num_threads == 0:
        err = "\nnum_threads cannot be zero. It can be -1 (use all available),\n"
        err += "or any integer > 0.\n\n"
        raise ValueError(err)


    # Figure out number of threads to use
    if num_threads < 0:
        try:
            num_threads = mp.cpu_count()
        except NotImplementedError:
            num_threads = os.cpu_count()
            if num_threads is None:
                print("Could not determine number of cpus. Using single thread.\n",flush=True)
                num_threads = 1

    # Figure out how many blocks we're going to download
    num_blocks = len(to_download) // block_size
    if len(to_download) % block_size > 0:
        num_blocks += 1
    print(f"Downloading {num_blocks} blocks of ~{block_size} sequences... ",flush=True)

    # queue will hold results from each download batch.
    manager = mp.Manager()
    queue = manager.Queue()
    lock = manager.Lock()
    with mp.Pool(num_threads) as pool:

        # This is a bit obscure. Build a list of args to pass to the pool. Each
        # tuple of args matches the args in _get_sequences_thread.
        # all_args has all len(df) recip blast runs we want to do.
        all_args = []
        for i in range(0,len(to_download),block_size):
            ids = ",".join(to_download[i:(i+block_size)])
            all_args.append((i,ids,num_tries_allowed,lock,queue))

        # Black magic. pool.imap() runs a function on elements in iterable,
        # filling threads as each job finishes. (Calls _get_sequences_thread
        # on every args tuple in all_args). tqdm gives us a status bar.
        # By wrapping pool.imap iterator in tqdm, we get a status bar that
        # updates as each thread finishes.
        list(tqdm(pool.imap(_get_sequences_thread,all_args),
                  total=len(all_args)))

    # Get results out of the queue. Sort by i to put in correct order
    results = []
    while not queue.empty():
        results.append(queue.get())
    results.sort()

    # Get final downloaded stuff
    total_download = []
    for r in results:
        total_download.append(r[1])

    print("Done.")

    return "".join(total_download)



"""
Generate ancestors and various summary outputs.
"""

import topiary

from ._raxml import run_raxml, RAXML_BINARY
from topiary.external._interface import create_new_dir, copy_input_file
from topiary.external._interface import prep_calc, write_run_information
from topiary import check

import pastml.acr
import ete3
from ete3 import Tree

import pandas as pd
import numpy as np

import os, re, glob, shutil

from matplotlib import pyplot as plt
import matplotlib.patches as patches
from matplotlib import gridspec

# -----------------------------------------------------------------------------
# Configure plotting
# -----------------------------------------------------------------------------

SMALL_SIZE = 14
MEDIUM_SIZE = 16
BIGGER_SIZE = 18

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

# -----------------------------------------------------------------------------
# Data
# -----------------------------------------------------------------------------

# lists of chemically similar amino acids. the same amino acid can occur
# in different lists.
CHEM_SIMILAR = [["A","C","N","Q","S","T"],
                ["A","V"],
                ["H","K","R"],
                ["D","E"],
                ["F","W","Y"],
                ["I","L","M","V"]]

# -----------------------------------------------------------------------------
# Ancestor processing functions
# -----------------------------------------------------------------------------

def _get_ancestral_gaps(alignment_file,tree_file):
    """
    Get ancestral gaps from an alignment and raxml output tree file. Gaps are
    reconstructed by parsimony using the DOWNPASS algorithm as implemented in
    pastml.

    Parameters
    ----------
        alignment_file: phy file used to generate ancestors in RAxML
        tree_file: output tree file with labeled internal nodes

    Return
    ------
        dictionary keying internal node names to lists of True (gap), False (no
        gap), and None (gapping unclear) for each site in that ancestor.
    """

    # Read the alignment file
    counter = 0
    leaf_names = []
    with open(alignment_file) as f:
        for line in f:

            # First line
            if counter == 0:
                col = line.split()
                num_taxa = int(col[0])
                num_sites = int(col[1])
                counter += 1

                char_matrix = np.zeros((num_taxa,num_sites),dtype=np.uint8)
                continue

            # Next, blank line
            if line.strip() == "":
                counter += 1
                continue

            # Alternating leaf id and sequence lines
            if counter % 2 == 0:
                leaf_names.append(line.strip())
                counter += 1
            else:
                index = (counter - 3)//2
                char_matrix[index,:] = np.array([c == "-" for c in line.strip()])
                counter += 1

    # Create a data frame where indexes are leaf names and columns are each gap
    out_dict = {}
    for i in range(char_matrix.shape[1]):
        out_dict[f"g{i}"] = char_matrix[:,i]
    gap_df = pd.DataFrame(out_dict)
    gap_df.index = leaf_names

    # Gaps, named as column names
    gap_names = list(gap_df.columns)

    # Load the tree, keeping the internal node names
    tree = Tree(tree_file,format=1)

    # Reconstruct gaps across tree by parsimony
    acr_result = pastml.acr.acr(tree,gap_df,prediction_method="DOWNPASS")

    # Create dictionary keying ancestor name to gap status across sequence
    gap_anc_dict = {}
    for node in tree.traverse('preorder'):
        if node.name in leaf_names:
            continue

        gap_anc_dict[node.name] = []

        # Go through gaps and grab from node feature
        for g in gap_names:
            state = node.__dict__[g]
            if len(state) == 1:
                if state == {0}:
                    state = False
                else:
                    state = True
            else:
                state = None


            gap_anc_dict[node.name].append(state)

    return gap_anc_dict

def _plot_ancestor_data(df_anc,
                        alt_anc_pp,
                        width_ratio,
                        anc_name,
                        anc_data_string):
    """
    Create a summary plot for an ancestor.

    Parameters
    ----------
        df_anc: ancestral data frame
        alt_anc_pp: cutoff (inclusive) for identifying plausible alternate states
        width_ratio: width ratio for plot
        anc_name: name of ancestor (name of plot, title on graph)
        anc_data_string: data to dump in subtitle

    Return
    ------
        None. writes pdf for ancestor
    """

    def _draw_histogram(values,ax,bin_size=0.05,color="gray"):
        """
        Draw a histogram sideways next to main plot.

        values: array to bin
        ax: axis on which to make plot
        bin_size: width of bins (from 0 to 1.0)
        color: color to make histogram bars

        returns maximum number of counts (for constructing xlim later)
        """

        # Create histogram
        counts, bins = np.histogram(values,
                                    bins=np.arange(0,1 + bin_size,bin_size))

        # Draw bars for histogram
        for i in range(len(counts)):
            rect = patches.Rectangle((0,bins[i]),
                                     width=counts[i],
                                     height=(bins[i+1]-bins[i]),
                                     linewidth=1,
                                     edgecolor='black',
                                     facecolor=color,
                                     alpha=0.5)
            ax.add_patch(rect)

        return np.max(counts)

    # Data frames containing unambiguous gaps and other sites
    df_gap = df_anc.loc[df_anc.site_type == "gap",:]
    df_nogap = df_anc.loc[df_anc.site_type != "gap",:]

    # create a figure
    fig = plt.figure()
    fig.set_figheight(4)
    fig.set_figwidth(10)

    # create grid for main and histogram subplots
    spec = gridspec.GridSpec(ncols=2, nrows=1,
                             width_ratios=[width_ratio, 1],
                             wspace=0.01)
    # Greate actual subplots
    ax = [fig.add_subplot(spec[0])]
    ax.append(fig.add_subplot(spec[1],sharey=ax[0]))

    # Plot gaps on main figure. Draw boxes for contiguous gap regions, Code
    # below is too clever by half, but returns contiguous blocks of gaps
    sites = np.array(df_gap.site,dtype=np.uint)
    contiguous = np.split(sites,np.where(np.diff(sites) != 1)[0]+1)

    # Iterate over contiguous blocks
    for c in contiguous:

        # Skip contigous gaps with no length.
        if len(c) == 0:
            continue

        # Find gap width.  Minimum gap width is 1.
        width = c[-1] - c[0]
        if width == 0:
            width = 1

        # Draw rectangle for gap block
        rect = patches.Rectangle((c[0],0),width,1,
                                 linewidth=1,
                                 edgecolor='lightgray',
                                 facecolor='lightgray')
        ax[0].add_patch(rect)

    # Create list of ambiguous gaps and draw veritical purple lines at these
    # positions.
    ambig_df = df_anc.loc[df_anc.site_type == "possible gap",:]
    for i in range(len(ambig_df)):
        row = ambig_df.iloc[i]
        ax[0].plot((row.site,row.site),(0,1.05),"--",lw=1,color="purple")

    # Plot ML and alt pp points
    ax[0].plot(df_nogap.site,df_nogap.ml_pp,".",color="black",markersize=8)
    ax[0].plot(df_nogap.site,df_nogap.alt_pp,".",color="red",markersize=8)

    # Plot ML and alt pp lines. Only draw lines over contiguous stretches.
    sites = np.array(df_nogap.site,dtype=np.uint)
    contiguous = np.split(sites,np.where(np.diff(sites) != 1)[0]+1)
    for c in contiguous:
        ax[0].plot(df_anc.site.iloc[c],df_anc.ml_pp.iloc[c],color="black",lw=2)
        ax[0].plot(df_anc.site.iloc[c],df_anc.alt_pp.iloc[c],color="red",lw=2)

    # Plot alt-all cutoff line
    ax[0].plot((np.min(df_anc.site),np.max(df_anc.site)),
               (alt_anc_pp,alt_anc_pp),"--",color="gray")

    # Draw histograms for ml and alt pp in right plot
    max_ml_counts = _draw_histogram(df_nogap.ml_pp,ax[1],color="gray")
    max_alt_counts = _draw_histogram(df_nogap.alt_pp,ax[1],color="red")

    # Figure out biggest value seen in histogram plot
    hist_x_max = np.max((max_ml_counts,max_alt_counts))

    # Plot alt-all cutoff line on histogram
    ax[1].plot((0,hist_x_max),
               (alt_anc_pp,alt_anc_pp),"--",color="gray")

    # Clean up axes for main plot
    ax[0].set_xlabel("alignment site")
    ax[0].set_ylabel("posterior probability")
    ax[0].set_xlim(np.min(df_anc.site),np.max(df_anc.site))
    ax[0].spines["right"].set_visible(False)
    ax[0].spines["top"].set_visible(False)
    ax[0].set_ylim(-0.05,1.1)

    # Clean up axes for histogram plot
    ax[1].set_xlim(0,hist_x_max*1.1)
    ax[1].spines["right"].set_visible(False)
    ax[1].spines["top"].set_visible(False)
    ax[1].spines["bottom"].set_visible(True)
    ax[1].spines["left"].set_visible(False)
    ax[1].axis('off')

    # Main plot title
    fig.suptitle(f"{anc_name}")

    # This bit of wackiness finds where on main plot the midpoint for the
    # whole graph is.
    main_plot_x_length = np.max(df_anc.site) - np.min(df_anc.site)
    total_plot_length = ((1 + width_ratio)/(width_ratio))*main_plot_x_length
    mid_point = total_plot_length/2.0

    # Plot the subtitle on the main plot
    ax[0].text(mid_point,1.08,anc_data_string,ha='center')

    # Save out figure
    fig.savefig(f"{anc_name}.pdf",bbox_inches="tight")


def _make_ancestor_summary_trees(df,
                                 avg_pp_dict,
                                 tree_file_with_labels,
                                 tree_file_with_supports=None):
    """
    Make trees summarizing ASR results. Creates three or four newick files:
        ancestors_label.newick: tree where internal names are labeled with
                                ancestor names
        ancestors_pp.newick: tree where supports are avg pp for that ancestor
        ancestors_support.newick: tree with supports (optional)
        ancestors_all.newick: tree where internal names are name|pp OR name|pp|support

    Parameters
    ----------
        df: topiary data frame
        avg_pp_dict: dictionary mapping ancestor names to avg ancestor posterior
                     probability
        tree_file_with_labels: output from RAxML that has nodes labeled by their
                               ancestor identity. Should also have branch lengths.
        tree_file_with_supports: tree file with supports (optional)

    Return
    ------
        None
    """

    # Create label trees
    t_labeled = Tree(tree_file_with_labels,format=1)

    # Create output trees
    t_out_label = Tree(tree_file_with_labels,format=1)
    t_out_pp = Tree(tree_file_with_labels,format=1)

    # Main iterator (over main labeled tree)
    input_label_iterator = t_labeled.traverse("preorder")

    # These sub iterators will get populated with final tree info
    label_iterator = t_out_label.traverse("preorder")
    pp_iterator = t_out_pp.traverse("preorder")

    if tree_file_with_supports is not None:
        t_out_supports = Tree(tree_file_with_supports,format=0)
        support_iterator = t_out_supports.traverse("preorder")

    # Iterate over main iterator
    for input_label_node in input_label_iterator:

        # Update sub itorators
        pp_node = next(pp_iterator)
        label_node = next(label_iterator)

        if tree_file_with_supports is not None:
            support_node = next(support_iterator)

        # See if this is a leaf and make sure the labeles
        is_label_leaf = input_label_node.is_leaf()
        is_pp_leaf = pp_node.is_leaf()

        # If at least one of the support or label nodes is a leaf...
        if is_label_leaf or is_pp_leaf:

            # If both are not leaves, this is bad news
            if not (is_label_leaf and is_pp_leaf):
                err = "Tree order mismatch (node type)\n"
                raise ValueError(err)
            else:

                # If the nodes are both leaves, make sure they have the
                # same name.
                if input_label_node.name != pp_node.name:
                    err = "Tree order mismatch (node name)\n"
                    err += f"{input_label_node.name} != {pp_node.name}\n"
                    raise ValueError(err)

            # If we get here, this is a leaf node. Continue the iteration.
            continue

        anc = input_label_node.name
        anc_name = re.sub("Node","anc",anc)

        label_node.name = anc_name
        pp_node.support = avg_pp_dict[anc_name]

    t_out_pp.write(format=2,format_root_node=True,outfile="ancestors_pp.newick")
    t_out_label.write(format=3,format_root_node=True,outfile="ancestors_label.newick")

    if tree_file_with_supports is not None:
        t_out_supports.write(format=3,format_root_node=True,outfile="ancestors_support.newick")


def _parse_raxml_anc_output(df,
                            anc_prob_file,
                            alignment_file,
                            tree_file_with_labels,
                            tree_file_with_supports=None,
                            dir_name="ancestors",
                            alt_cutoff=0.25,
                            plot_width_ratio=5):
    """
    Parse raxml marginal ancestral state reconstruction output and put out in
    human-readable fashion. Writes fasta file and csv file with ancestors.
    Writes final newick with three trees: ancestor label, posterior probability,
    and SH support. Creates creates summary plots for all ancestors.

    Parameters
    ----------
        df: topiary data frame
        anc_prob_file: ancestor posterior probability file as written out by raxml
        alignment_file: phylip alignment used to create ancestors
        tree_file_with_labels: output newick tree file written by raxml
        tree_file_with_supports: newick tree with supports (optional)
        name: name for output directory
        alt_cutoff: cutoff (inclusive) for identifying plausible alternate states
        plot_width_ratio: ratio of main and histogram plot widths for ancestors

    Return
    ------
        None
    """

    # Make directory and copy in files
    dir_name = create_new_dir(dir_name)

    anc_prob_file = copy_input_file(anc_prob_file,dir_name)
    alignment_file = copy_input_file(alignment_file,dir_name)
    tree_file_with_labels = copy_input_file(tree_file_with_labels,dir_name)
    if tree_file_with_supports is not None:
        tree_file_with_supports = copy_input_file(tree_file_with_supports,
                                                  dir_name)

    # Move into output directory
    cwd = os.getcwd()
    os.chdir(dir_name)

    # Get gaps, reconstructed by parsimony
    gap_anc_dict = _get_ancestral_gaps(alignment_file,
                                       tree_file_with_labels)

    # Read pp file into a list of ancestors (anc_list) and posterior
    # probabilities for amino acids at each site
    anc_list = []
    anc_all_pp = []
    last_line = ""
    first_line = True
    with open(anc_prob_file) as f:
        for l in f:
            line = l.strip()

            if first_line:
                tmp_column_names = line.split()
                column_names = []
                for c in tmp_column_names:
                    column_names.append(re.sub("p_","",c))

                aa_list = column_names[3:]

                first_line = False
                continue

            col = line.split()

            if len(anc_list) == 0:
                anc_list.append(col[0])
                anc_all_pp.append([])

            if col[0] != anc_list[-1]:
                anc_list.append(col[0])
                anc_all_pp.append([])

            anc_all_pp[-1].append(np.array([float(c) for c in col[3:]]))


    # Create data structures for output
    out = []
    df_list = []
    avg_pp_dict = {}

    # Go through each ancestor
    for i in range(len(anc_list)):

        for_df = {"anc":[],"site":[],"gap":[],
                  "ml_state":[],"ml_pp":[],
                  "alt_state":[],"alt_pp":[],
                  "site_type":[],"entropy":[]}

        anc = anc_list[i]
        anc_name = re.sub("Node","anc",anc)

        # Get ml and next best seq and posterior probability. Entropy holds
        # the shannon entropy for all posterior probabilities across the
        # row.  If entropy is very high, we have very little confidence --
        # likely a gap.
        anc_ml_seq = []
        anc_ml_pp = []
        anc_alt_seq = []
        anc_alt_pp = []
        entropy = []

        # Go through sites
        for j in range(len(anc_all_pp[i])):

            # Get posterior probability vector for site
            pp = anc_all_pp[i][j].copy()

            # Get entropy, ignoring positions with pp = 0
            tmp_pp = pp[pp != 0.0]
            entropy.append(np.sum(-tmp_pp*np.log(tmp_pp)))

            # Get ml posterior probability and sequence
            anc_ml_pp.append(np.max(pp))
            anc_ml_seq.append(aa_list[np.argmax(pp)])

            # Set max to zero to get next best
            pp[np.argmax(pp)] = 0.0

            # Get second best posterior probability and sequence
            anc_alt_pp.append(np.max(pp))
            anc_alt_seq.append(aa_list[np.argmax(pp)])

            # Add gap!
            if gap_anc_dict[anc][j] == True:
                anc_ml_pp[-1] = 1.0
                anc_ml_seq[-1] = "-"
                anc_alt_pp[-1] = 0.0
                anc_alt_seq[-1] = "-"

        # Convert lists read into numpy arrays
        ml_pp = np.array(anc_ml_pp)
        ml_seq = np.array(anc_ml_seq)
        alt_pp = np.array(anc_alt_pp)
        alt_seq = np.array(anc_alt_seq)

        # Create alt all sequence. This sequence is ML at all states except
        # those with pp >= some alternate cutoff (usually 0.25)
        alt_mask = alt_pp >= alt_cutoff
        alt_all_seq = ml_seq.copy()
        alt_all_seq[alt_mask] = alt_seq[alt_mask]
        alt_all_pp = ml_pp.copy()
        alt_all_pp[alt_mask] = alt_pp[alt_mask]

        # Write out information for data frame
        for_df["anc"] = [anc_name for _ in range(len(anc_all_pp[i]))]
        for_df["site"] = [j for j in range(len(anc_all_pp[i]))]
        for_df["gap"] = gap_anc_dict[anc]

        for_df["ml_state"] = list(ml_seq)
        for_df["ml_pp"] = list(ml_pp)
        for_df["alt_state"] = list(alt_seq)
        for_df["alt_pp"] = list(alt_pp)
        for_df["entropy"] = list(entropy)

        # Classify sites according to semi-intelligent criteria. If gapping
        # unclear by parsimony, call as "possible gap."  If it is a gap, call
        # as "gap." If second best pp is above cutoff, call
        # ambiguous.  Look at ml and alt aa to see if this is a chemically
        # similar or dissimilar ambiguity.  Otherwise, call 'good'
        site_type = []
        num_ambig_gaps = 0
        for j in range(len(anc_all_pp[i])):

            if gap_anc_dict[anc][j] is None:
                site_type.append("possible gap")
                num_ambig_gaps += 1

            elif gap_anc_dict[anc][j]:
                site_type.append("gap")

            elif alt_pp[j] >= alt_cutoff:

                found_match = False
                for ml_set in CHEM_SIMILAR:
                    if ml_seq[j] in ml_set:
                        if alt_seq[j] in ml_set:
                            found_match = True
                            continue
                if found_match:
                    site_type.append("ambig_similar")
                else:
                    site_type.append("ambig_dissimilar")
            else:
                site_type.append("good")

        for_df["site_type"] = site_type

        df_list.append(pd.DataFrame(for_df))

        # Calculate average pp, log sum pp, and number of ambiguous sites
        ml_avgPP = np.mean(ml_pp)
        ml_lnPP = np.sum(np.log(ml_pp))
        alt_avgPP = np.mean(alt_all_pp)
        alt_lnPP = np.sum(np.log(alt_all_pp))
        num_ambig = np.sum(alt_mask)

        # Record avg pp
        avg_pp_dict[anc_name] = ml_avgPP

        # Write ML sequence to fasta
        header = [f">{anc_name}",
                  f"avgPP:{ml_avgPP:.3f}",f"lnPP:{ml_lnPP:.3f}",
                  f"num_ambig:{num_ambig}",f"num_ambig_gaps:{num_ambig_gaps}\n"]
        out.append("|".join(header))
        out.append("".join(ml_seq))
        out.append("\n")

        # Write alt sequence to fasta
        header = [f">{anc_name}_altAll",
                  f"avgPP:{alt_avgPP:.3f}",f"lnPP:{alt_lnPP:.3f}",
                  f"num_ambig:{num_ambig}",f"num_ambig_gaps:{num_ambig_gaps}\n"]
        out.append("|".join(header))
        out.append("".join(alt_all_seq))
        out.append("\n")

        # String for subtitle
        anc_data_string = [f"avgPP: {ml_avgPP:.3f}",
                           f"lnPP: {ml_lnPP:.0f}",
                           f"# ambig: {num_ambig}",
                           f"# ambig gaps: {num_ambig_gaps}"]

        # Plot a summary of the ancestor
        _plot_ancestor_data(df_list[-1],
                            alt_anc_pp=alt_cutoff,
                            width_ratio=plot_width_ratio,
                            anc_name=anc_name,
                            anc_data_string=", ".join(anc_data_string))


    # Write final fasta file
    f = open(f"ancestors.fasta","w")
    f.write("".join(out))
    f.close()

    # Write ancestor df to csv
    anc_df = pd.concat(df_list,ignore_index=True)
    anc_df.to_csv(f"ancestors.csv")

    # Create final tree
    _make_ancestor_summary_trees(df,
                                 avg_pp_dict,
                                 tree_file_with_labels,
                                 tree_file_with_supports)

    os.chdir(cwd)


def generate_ancestors(previous_dir=None,
                       df=None,
                       model=None,
                       tree_file=None,
                       tree_file_with_supports=None,
                       alt_cutoff=0.25,
                       output=None,
                       overwrite=False,
                       threads=-1,
                       raxml_binary=RAXML_BINARY):
    """
    Generate ancestors and various summary outputs. Creates fasta file and csv
    file with ancestral sequences, set of ancestor plots, and a tree with
    ancestral names and supports.

    Parameters
    ----------
    previous_dir : str, optional
        directory containing previous calculation. function will grab the the
        csv, model, and tree from the previous run. If this is not specified,
        `df`, `model`, and `tree_file` arguments must be specified.
    df : pandas.DataFrame or str, optional
        topiary data frame or csv written out from topiary df. Will override
        dataframe from `previous_dir` if specified.
    model : str, optional
        model (i.e. "LG+G8"). Will override model from `previous_dir`
        if specified.
    tree_file : str
        tree_file in newick format. Will override tree from `previous_dir` if
        specified.
    tree_file_with_supports : str
        tree file with supports for each node in newick format. Will override
        tree_file_with_supports from `previous_dir` if specified.
    alt_cutoff : float, default=0.25
        cutoff to use for altAll calculation. Should be between 0 and 1.
    output : str, optional
        output directory. If not specified, create an output directory
        with form "generate_ancestors_randomletters"
    overwrite : bool, default=False
        whether or not to overwrite existing output
    threads : int, default=-1
        number of threads to use. if -1, use all avaialable
    raxml_binary : str, optional
        what raxml binary to use

    Returns
    -------
    Python.core.display.Image or None
        if running in jupyter notebook, return Image of tree with ancestors
        drawn; otherwise, return None
    """

    result = prep_calc(previous_dir=previous_dir,
                       df=df,
                       model=model,
                       tree_file=tree_file,
                       other_files=[tree_file_with_supports],
                       output=output,
                       overwrite=overwrite,
                       output_base="generate_ancestors")

    df = result["df"]
    csv_file = result["csv_file"]
    model = result["model"]
    tree_file = result["tree_file"]
    alignment_file = result["alignment_file"]
    tree_file_with_supports = result["other_files"][0]
    starting_dir = result["starting_dir"]
    output = result["output"]

    alt_cutoff = check.check_float(alt_cutoff,
                                   "alt_cutoff",
                                   minimum_allowed=0,
                                   maximum_allowed=1)

    # Do marginal reconstruction on the tree
    cmd = run_raxml(algorithm="--ancestral",
                    alignment_file=alignment_file,
                    tree_file=tree_file,
                    model=model,
                    seed=True,
                    dir_name="working_inference",
                    threads=threads,
                    raxml_binary=raxml_binary)

    anc_prob_file = os.path.join("working_inference",
                                 "alignment.phy.raxml.ancestralProbs")
    tree_file_with_labels = os.path.join("working_inference",
                                         "alignment.phy.raxml.ancestralTree")

    # Parse output and make something human-readable
    _parse_raxml_anc_output(df,
                            anc_prob_file,
                            alignment_file,
                            tree_file_with_labels,
                            tree_file_with_supports,
                            dir_name="working_analysis",
                            alt_cutoff=alt_cutoff)

    outdir = "output"
    os.mkdir(outdir)

    # tree file
    shutil.copy(tree_file,os.path.join(outdir,"tree.newick"))

    # Write run information
    write_run_information(outdir=outdir,
                          df=df,
                          calc_type="ancestors",
                          model=model,
                          cmd=cmd,
                          outgroup=result["outgroup"])

    # Copy ancestor files into an ancestors directory
    files_to_grab = glob.glob(os.path.join("working_analysis","*.*"))
    anc_out = os.path.join(outdir,"ancestors")
    os.mkdir(anc_out)
    for f in files_to_grab:
        shutil.copy(f,os.path.join(anc_out,os.path.split(f)[-1]))

    print(f"\nWrote results to {os.path.abspath(outdir)}\n")

    # Leave working directory
    os.chdir(starting_dir)

    # Create a plot of the tree
    ret = topiary.draw.ancestor_tree(run_dir=output,
                                     output_file=os.path.join(output,
                                                              "output",
                                                              "summary-tree.pdf"))
    if topiary._in_notebook:
        return ret



"""
Use entrez to download a proteome from the NCBI.
"""

import topiary
from topiary import check

from Bio import Entrez

import os, urllib, datetime, re

def _get_genome_url(record):
    """
    Get the url pointing to a genome with information for selecting the best
    genome.

    If a url is found, returns a tuple where entries are:

        RefSeq_category:
            0 -> reference genome
            1 -> representative genome
            2 -> something else
        Url class:
            0 -> FtpPath_RefSeq
            1 -> FtpPath_GenBank
        LastUpdateDate (as datetime.datetime inference)
        url

    If one sorts a list of these tuples, the first entry will be the best
    genome. If no genome is found, function returns None.

    Parameters
    ----------
        record: esummary record returned by Entrez.read

    Return
    ------
        genome info or None
    """

    raw_date = re.sub("/","-",record["LastUpdateDate"])
    last_date = datetime.datetime.fromisoformat(raw_date)

    if record["RefSeq_category"] == "reference genome":

        for i, r in enumerate(["FtpPath_RefSeq","FtpPath_GenBank"]):
            if record[r] != "":
                this_url = str(record[r])
                return (0,i,last_date,this_url)

    elif record["RefSeq_category"] == "representative genome":

        for i, r in enumerate(["FtpPath_RefSeq","FtpPath_GenBank"]):
            if record[r] != "":
                this_url = str(record[r])
                return (1,i,last_date,this_url)

    else:

        for i, r in enumerate(["FtpPath_RefSeq","FtpPath_GenBank"]):
            if record[r] != "":
                this_url = str(record[r])
                return (2,i,last_date,this_url)

    return None

def get_proteome(taxid=None,species=None,output_dir="."):
    """
    Use entrez to download a proteome from the NCBI.

    Parameters
    ----------
    taxid: int or str, optional
        NCBI taxid (integer or string version of the integer). Incompatible with
        `species` argument. At least taxid or species must be specified.
    species : str, optional
        bionomial name of species (i.e. Mus musculus). Incompatible with `taxid`
        argument. At least taxid or species must be specified.
    output_dir : str, optional
        where to write the file locally

    Returns
    -------
    proteome_file : str or None
        the file we downloaded or None if no file downloaded
    """

    if species is None and taxid is None:
        err = "\nYou must specify either species or taxid, but not both.\n\n"
        raise ValueError(err)

    if species is not None and taxid is not None:
        err = "\nYou must specify either species or taxid, but not both.\n\n"
        raise ValueError(err)

    output_dir = str(output_dir)
    if not os.path.isdir(output_dir) or not os.access(output_dir,os.W_OK):
        err = "\nlocal_path must exist and be writable.\n\n"
        raise ValueError(err)

    # Do query using taxid. This will throw an error if species is not sane,
    # so is a helpful check before querying ncbi.
    if species:
        print(f"Downloading proteome for species '{species}'",flush=True)
        taxid = topiary.ncbi.get_taxid(species)
    else:
        print(f"Downloading proteome for taxid '{taxid}'",flush=True)

    # Make sure the taxid is sane
    taxid = check.check_int(taxid,"taxid")

    # Look for assembly ids for this taxid
    query_text = f"(txid{taxid}[ORGN])"
    esearch_handle = Entrez.esearch(db="assembly",
                                    retmax=1000,
                                    term=query_text,
                                    idtype="acc")
    search_record = Entrez.read(esearch_handle)

    # Get assembly ids
    try:
        returned_ids = list(search_record["IdList"])
    except KeyError:
        err = "\nThe Entrez.esearch query failed.\n\n"
        raise RuntimeError(err)

    # Make sure something came back
    if len(returned_ids) == 0:
        err = f"\nThe query '{query_text}' returned no assemblies.\n\n"
        raise RuntimeError(err)

    # Now get summary data for these records.
    esummary_query = ",".join(returned_ids)
    esummary_handle = Entrez.esummary(db="assembly",id=esummary_query)
    esummary_record = Entrez.read(esummary_handle)

    try:
        records = esummary_record["DocumentSummarySet"]["DocumentSummary"]
    except KeyError:
        err = "\nThe Entrez.esummary query failed.\n\n"
        raise RuntimeError(err)

    # Get list of urls from all records
    urls = []
    for record in records:
        ret = _get_genome_url(record)
        if ret is not None:
            urls.append(ret)
    urls.sort()

    # Go through the urls from highest to lowest quality and try to download
    # the proteome.
    success = False
    for u in urls:

        try:
            genome_url = u[3]
            refseq_name = os.path.basename(genome_url)
            out_file = f"{refseq_name}_protein.faa.gz"
            remote_file = f"{genome_url}/{out_file}"
            local_file = os.path.join(output_dir,out_file)
            urllib.request.urlretrieve(remote_file, local_file)
            success = True
            break

        except (urllib.error.URLError,urllib.error.HTTPError):
            continue

    # Try to delete random file that gets downloaded when we make this query.
    try:
        os.remove(os.path.join(output_dir,"esummary_assembly.dtd"))
    except FileNotFoundError:
        pass


    if success:
        return local_file

    return None



"""
Interfaces to external programs: raxml-ng, generax, muscle, ncbi, and opentree.
"""

from . import generax
from . import muscle
from . import ncbi
from . import opentree
from . import raxml
from . import _interface

from .opentree import get_ott_id, get_species_tree
from .muscle import run_muscle
from .ncbi.blast import recip_blast

from .raxml import find_best_model
from .raxml import generate_ml_tree
from .raxml import generate_ancestors

from .generax import reconcile




import topiary
from topiary import check
import pandas as pd
import numpy as np

import os, re, shutil

from .base import _run_and_print


def read_paralog_patterns(paralog_patterns):

    err_summary = ["This file should have the following format:\n",
                   "paralog1;pattern1;pattern2;pattern3"
                   "paralog2;pattern4;pattern5;pattern6",
                   "...\n",
                   "For a real pair of proteins, this could be:\n",
                   "LY96;MD2;MD-2;ESOP1",
                   "LY86;MD1;MD-1\n",
                   "Blank lines and anything after # on a line are ignored\n"]
    err_summary = "\n".join(err_summary)

    if type(paralog_patterns) is str:
        if os.path.isfile(paralog_patterns):

            out_dict = {}
            with open(paralog_patterns) as f:
                for line in f:

                    # Remove everything after '#' (comment)
                    l = line.split("#")[0].strip()

                    # Skip blank lines
                    if l.strip() == "":
                        continue

                    # Now split line on ";" and get non-"" entries.
                    col = [c.strip() for c in l.split(";")]
                    col = [c for c in col if c != ""]

                    if len(col) < 2:
                        err = f"\nCould not parse line '{line}' in paralog_patterns.\n\n"
                        err += err_summary
                        raise ValueError(err)

                    # Create pattern dictionary, searching for all unique patterns
                    out_dict[col[0]] = list(set(col))

            paralog_patterns = out_dict

    patterns = check.check_paralog_patterns(paralog_patterns)

    return dict([(p[1],p[0]) for p in patterns])


def read_key_species(key_species):
    """
    """

    err_summary = ["This file should have the following format:\n",
                   "species 1;species 2;species 3; ...\n\n"
                   "For example, this could be:\n",
                   "Homo sapiens;Mus musculus;Danio rerio\n\n",
                   "Blank lines and anything after # on a line are ignored\n"]
    err_summary = "\n".join(err_summary)

    if type(key_species) is str:
        if os.path.isfile(key_species):

            out_list = []
            with open(key_species) as f:
                for line in f:

                    # Remove everything after '#' (comment)
                    l = line.split("#")[0].strip()

                    # Skip blank lines
                    if l.strip() == "":
                        continue

                    # Now split line on ";" and get non-"" entries.
                    col = [c.strip() for c in l.split(";")]
                    col = [c for c in col if c != ""]

                    out_list.extend(col)

            if len(out_list) == 0:
                w = f"\nWarning: no key species found in '{key_species}'\n"
                print(w,flush=True)

            key_species = out_list[:]

    key_species = check.check_iter(key_species,
                                               "key_species",
                                               required_value_type=str,
                                               is_not_type=[str,dict])

    return key_species

def read_ncbi_taxid(ncbi_rev_blast_taxid):

    taxid_out = []
    taxid_is_bad = True

    # Is it int-like? Convert integer input to list of one string. We're
    # stringent on this -- no floats -- because an int cast will silently
    # round down. If someone used a taxid like 960.6 (extra .) it would be
    # interpreted as 960, which is very much the wrong taxid
    if np.issubdtype(type(ncbi_rev_blast_taxid),np.integer):
        ncbi_rev_blast_taxid = [f"{ncbi_rev_blast_taxid}"]

    # This wacky line sees if the taxid is iterable, but not a type *class*.
    # Catches weird edge case where user passes in str or int as a taxid
    if hasattr(ncbi_rev_blast_taxid,"__iter__") and type(ncbi_rev_blast_taxid) is not type:

        taxid_is_bad = False

        # If taxid is a single string, convert it to a list of one string. Split
        # on "," in case coming in from command line.
        if type(ncbi_rev_blast_taxid) is str:
            ncbi_rev_blast_taxid = [t.strip() for t in ncbi_rev_blast_taxid.split(",")]

        # Go through list of taxids and put in correct format
        for t in ncbi_rev_blast_taxid:
            if type(t) is str:
                taxid_out.append(f"txid{t}[ORGN]")
            elif np.issubdtype(type(t),np.integer):
                taxid_out.append(f"txid{t}[ORGN]")
            else:
                taxid_is_bad = True
                break

    # If taxid was None to begin with, ignore it
    else:
        if ncbi_rev_blast_taxid is None:
            taxid_out = []
            taxid_is_bad = False

    if taxid_is_bad:
        err = "\ntaxid should be either a single ncbi taxon id (e.g. 9606 for\n"
        err += "human) or list of ncbi taxon ids. Individual ids can either be\n"
        err += "int or str. These are passed to NCBI without further validation.\n\n"
        raise ValueError(err)


def rockit(xml_input,
           paralog_patterns,
           key_species=[],
           output_dir="rockit",
           min_redundancy_cutoff=0.85,
           target_number_of_sequences=500,
           ncbi_blast_db="nr",
           local_blast_db=None,
           ncbi_rev_blast_taxid="9606",
           phylo_context="All life",
           clean_alignment=True,
           overwrite=False):
    """

    Parameters
    ----------
        xml_input: xml file, directory containing xml files, or list of xml files

        paralog_patterns: dictionary of paralog patterns or file containing
                          paralog_patterns

        key_species: list of key species or file containing key species

        output_dir: output directory

        min_redundancy_cutoff: when selecting a redundancy cutoff, do not select
                               a cutoff less than this

        target_number_of_sequences: find a redundancy cutoff that yields
                                    approximately this number of sequences in
                                    final alignment

        ncbi_blast_db: NCBI recip blast database. Ignored if local_blast_db
                           is specified

        local_blast_db: local recip blast database to use.

        ncbi_rev_blast_taxid: limit NCBI recip blast to the specified taxid.
                              integer or list of integers. If called from the
                              command line use the format:
                              --ncbi_blast_db=9606,10090

        phylo_context: look for species within a specific phylogenetic context
                       on open tree of life. To see what's available given the
                       current opentree database, type the following on your
                       python terminal:

                            from opentree import OT
                            print(OT.tnrs_contexts().response_dict)

        clean_alignment: whether or not to try to clean up the alignment to
                         remove sparse/gappy sequences.

        overwrite: overwrite outputs. True or False. On command line, use
                    --overwrite.

    Return
    ------
        None.

        Creates and populates output_dir with output files. Key outputs are
        final-dataframe.csv and final-alignment.fasta. All files start with
        pipeline step number.
    """

    # Get current workign directory
    current_dir = os.getcwd()

    # See if this is an input directory. If so, make it a full absolute path
    if os.path.exists(str(xml_input)):
        xml_input = os.path.abspath(str(xml_input))

    # Deal with the paralog_patterns argument (either read file or just load dict)
    paralog_patterns = read_paralog_patterns(paralog_patterns)

    # Deal with the key_species argument (either read file or just load dict)
    key_species = read_key_species(key_species)

    # check min_redundancy_cutoff
    min_redundancy_cutoff = check.check_float(min_redundancy_cutoff,
                                                          "min_redundancy_cutoff",
                                                          minimum_allowed=0,
                                                          maximum_allowed=1)

    # check target_number_of_sequences
    target_number_of_sequences = check.check_int(target_number_of_sequences,
                                                             "target_number_of_sequences",
                                                             minimum_allowed=1)

    # If local_blast_db is not None, create absolute path to that file
    if local_blast_db is not None:
        local_blast_db = str(local_blast_db)
        local_blast_db = os.path.join(current_dir,local_blast_db)

        # Check for existance of blast database
        blast_file = f"{local_blast_db}.psq"
        if not os.path.isfile(blast_file):
            err = f"\ncould not find local blast database {local_blast_db}\n\n"
            raise ValueError(err)

        # Set ncbi recip blast to None -- override with local
        ncbi_blast_db = None

    # Make ncbi_blast_db a string
    if ncbi_blast_db is not None:
        ncbi_blast_db = str(ncbi_blast_db)
        ncbi_rev_blast_taxid = read_ncbi_taxid(ncbi_rev_blast_taxid)

    # Make sure the phylogenetic context is sane
    phylo_context = topiary.opentree.is_allowed_phylo_context(phylo_context)

    # Check for existance of output_directory
    output_dir = str(output_dir)
    if os.path.isdir(output_dir):
        if overwrite:
            shutil.rmtree(output_dir)
        else:
            err = f"\noutput directory '{output_dir}' already exists.\n"
            err += "To overwrite, set overwrite = True.\n\n"
            raise FileExistsError(err)

    # Move into output directory.
    current_dir = os.getcwd()
    os.mkdir(output_dir)
    os.chdir(output_dir)

    step_counter = 0

    # Load xml files and download sequences
    df, step_counter = _run_and_print(function=topiary.df_from_blast_xml,
                                      kwargs={"xml_input":xml_input},
                                      step_counter=step_counter,
                                      out_file_string="initial",
                                      human_string="Loading xml files.")

    # Make human readable sequence nicknames
    df, step_counter = _run_and_print(function=topiary.create_nicknames,
                                      kwargs={"df":df,"paralog_patterns":paralog_patterns},
                                      step_counter=step_counter,
                                      out_file_string="nickname",
                                      human_string="Assigning sequence nicknames.")

    # Get ott id. We do this early on in case we can't find a species ott. This
    # will set those sequences to False
    df, step_counter = _run_and_print(function=topiary.get_ott_id,
                                      kwargs={"df":df,"phylo_context":phylo_context},
                                      step_counter=step_counter,
                                      out_file_string="ott",
                                      human_string="Getting OTT (species) ids.")


    # Remove basically identical sequences early on. Goal is to minimize
    # sequences passed to recip blast.
    df, step_counter = _run_and_print(function=topiary.remove_redundancy,
                                      kwargs={"df":df,"cutoff":0.999,"key_species":key_species,"only_in_species":True,"silent":False},
                                      step_counter=step_counter,
                                      out_file_string="remove-identical",
                                      human_string="Removing identical sequences within species.")


    # If a recip blast database was passed in, do recip blast.
    if local_blast_db is not None or ncbi_blast_db is not None:

        kwargs = {"df":df,
                  "paralog_patterns":paralog_patterns,
                  "local_blast_db":local_blast_db,
                  "ncbi_blast_db":ncbi_blast_db,
                  "ncbi_taxid":ncbi_rev_blast_taxid}
        df, step_counter = _run_and_print(function=topiary.recip_blast,
                                          kwargs=kwargs,
                                          step_counter=step_counter,
                                          out_file_string="recip-blast",
                                          human_string="Running recip blast.")

    # Identify a redundancy cutoff that yields approximately the target number
    # of sequences.
    print("-------------------------------------------------------------------")
    print(f"Identifying redundancy cutoff that yields ~ {target_number_of_sequences} sequences.")
    print("-------------------------------------------------------------------")
    print("",flush=True)
    cutoff = topiary.quality.find_cutoff(df,
                                         min_cutoff=min_redundancy_cutoff,
                                         target_number=target_number_of_sequences)
    print()
    print(f"Found cutoff: {cutoff:.2f}\n",flush=True)

    # Lower sequence redundancy
    df, step_counter = _run_and_print(function=topiary.remove_redundancy,
                                      kwargs={"df":df,"cutoff":cutoff,"key_species":key_species},
                                      step_counter=step_counter,
                                      out_file_string="lower-redundancy",
                                      human_string="Lowering sequence redundancy.")



    # Clean alignment
    if clean_alignment:

        df, step_counter = _run_and_print(function=topiary.run_muscle,
                                          kwargs={"input":df},
                                          step_counter=step_counter,
                                          out_file_string="first-pass-alignment",
                                          human_string="Doing initial alignment.")

        df, step_counter = _run_and_print(function=topiary.quality.clean_alignment,
                                          kwargs={"df":df,"key_species":key_species},
                                          step_counter=step_counter,
                                          out_file_string="cleaned-up-alignment",
                                          human_string="Removing difficult-to-align sequences.")

    # Align final set of high quality sequences
    df, step_counter = _run_and_print(function=topiary.run_muscle,
                                      kwargs={"input":df},
                                      step_counter=step_counter,
                                      out_file_string="final-alignment",
                                      human_string="Doing final alignment.")

    alignment_file = f"{step_counter:02d}_final-alignment.fasta"

    print("-------------------------------------------------------------------")
    print(f"Writing final alignment to {os.path.join(output_dir,alignment_file)}")
    print("-------------------------------------------------------------------")
    print("",flush=True)

    topiary.write_fasta(df,alignment_file,
                        seq_column="alignment")

    os.chdir(current_dir)
